#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Given a TBW file, plot the time averaged spectra for each digitizer input."""

import os
import sys
import math
import numpy
import ephem
import argparse

from lsl.common import stations, metabundle
from lsl.reader.ldp import LWA1DataFile
from lsl.correlator import fx as fxc
from lsl.astro import unix_to_utcjd, DJD_OFFSET
from lsl.misc import parser as aph

import matplotlib.pyplot as plt


def main(args):
    # Setup the LWA station information
    if args.metadata is not None:
        try:
            station = stations.parseSSMIF(args.metadata)
        except ValueError:
            station = metabundle.getStation(args.metadata, ApplySDM=True)
    else:
        station = stations.lwa1
    antennas = station.getAntennas()
    
    # Length of the FFT
    LFFT = args.fft_length
    
    # Make sure that the file chunk size contains is an integer multiple
    # of the FFT length so that no data gets dropped
    maxFrames = int((30000*260)/float(LFFT))*LFFT
    # It seems like that would be a good idea, however...  TBW data comes one
    # capture at a time so doing something like this actually truncates data 
    # from the last set of stands for the first integration.  So, we really 
    # should stick with
    maxFrames = (30000*260)
    
    idf = LWA1DataFile(args.filename)
    
    nFrames = idf.getInfo('nFrames')
    srate = idf.getInfo('sampleRate')
    dataBits = idf.getInfo('dataBits')
    # The number of ant/pols in the file is hard coded because I cannot figure out 
    # a way to get this number in a systematic fashion
    antpols = len(antennas)
    nChunks = int(math.ceil(1.0*nFrames/maxFrames))
    if dataBits == 12:
        nSamples = 400
    else:
        nSamples = 1200
        
    # Read in the first frame and get the date/time of the first sample 
    # of the frame.  This is needed to get the list of stands.
    beginDate = ephem.Date(unix_to_utcjd(idf.getInfo('tStart')) - DJD_OFFSET)
    
    # File summary
    print "Filename: %s" % args.filename
    print "Date of First Frame: %s" % str(beginDate)
    print "Ant/Pols: %i" % antpols
    print "Sample Length: %i-bit" % dataBits
    print "Frames: %i" % nFrames
    print "Chunks: %i" % nChunks
    print "==="
    
    # Setup the window function to use
    if args.bartlett:
        window = numpy.bartlett
    elif args.blackman:
        window = numpy.blackman
    elif args.hanning:
        window = numpy.hanning
    else:
        window = fxc.noWindow
        
    # Master loop over all of the file chunks
    nChunks = 1
    masterSpectra = numpy.zeros((nChunks, antpols, LFFT))
    masterWeight = numpy.zeros((nChunks, antpols, LFFT))
    
    readT, t, data = idf.read(0.061)
    
    # Calculate the spectra for this block of data and then weight the results by 
    # the total number of frames read.  This is needed to keep the averages correct.
    # NB:  The weighting is the same for the x and y polarizations because of how 
    # the data are packed in TBW
    freq, tempSpec = fxc.SpecMaster(data, LFFT=LFFT, window=window, verbose=args.verbose)
    for stand in xrange(masterSpectra.shape[1]):
        masterSpectra[0,stand,:] = tempSpec[stand,:]
        masterWeight[0,stand,:] = int(readT*srate/LFFT)
        
    # We don't really need the data array anymore, so delete it
    del(data)
    
    # Apply the cable loss corrections, if requested
    if args.gain_correct:
        for s in xrange(masterSpectra.shape[1]):
            currGain = antennas[s].cable.gain(freq)
            for c in xrange(masterSpectra.shape[0]):
                masterSpectra[c,s,:] /= currGain
                
    # Now that we have read through all of the chunks, perform the final averaging by
    # dividing by all of the chunks
    spec = masterSpectra.mean(axis=0)
    
    # The plots:  This is setup for the current configuration of 20 antpols
    if args.gain_correct & args.stack:
        # Stacked spectra - only if cable loss corrections are to be applied
        colors = ['blue', 'green', 'red', 'cyan', 'magenta', 'black', 
            'purple', 'salmon', 'olive', 'maroon', 'saddlebrown', 'yellowgreen', 
            'teal', 'steelblue', 'seagreen', 'slategray', 'mediumorchid', 'lime', 
            'dodgerblue', 'darkorange']
            
        for f in xrange(int(numpy.ceil(antpols/20.))):
            fig = plt.figure()
            ax1 = fig.add_subplot(1, 1, 1)
            for i in xrange(f*20, f*20+20):
                currSpectra = numpy.squeeze( numpy.log10(spec[i,:])*10.0 )
                ax1.plot(freq/1e6, currSpectra, label='%i,%i' % (antennas[i].stand.id, antennas[i].pol), color=colors[i % 20])
                
            ax1.set_xlabel('Frequency [MHz]')
            ax1.set_ylabel('P.S.D. [dB/RBW]')
            ax1.set_xlim([20,88])
            #ax1.set_ylim([10,90])
            leg = ax1.legend(loc=0, ncol=3)
            for l in leg.get_lines():
                l.set_linewidth(1.7)  # the legend line width
    else:
        for f in xrange(int(numpy.ceil(antpols/20))):
            # Normal plotting
            fig = plt.figure()
            figsY = 4
            figsX = 5
            fig.subplots_adjust(left=0.06, bottom=0.06, right=0.94, top=0.94, wspace=0.20, hspace=0.50)
            for i in xrange(f*20, f*20+20):
                ax = fig.add_subplot(figsX, figsY, (i%20)+1)
                try:
                    currSpectra = numpy.squeeze( numpy.log10(spec[i,:])*10.0 )
                except IndexError:
                    break
                ax.plot(freq/1e6, currSpectra, label='Stand: %i, Pol: %i (Dig: %i)' % (antennas[i].stand.id, antennas[i].pol, antennas[i].digitizer))
                
                # If there is more than one chunk, plot the difference between the global 
                # average and each chunk
                if nChunks > 1:
                    for j in xrange(nChunks):
                        # Some files are padded by zeros at the end and, thus, carry no 
                        # weight in the average spectra.  Skip over those.
                        if masterWeight[j,i,:].sum() == 0:
                            continue
                            
                        # Calculate the difference between the spectra and plot
                        subspectra = numpy.squeeze( numpy.log10(masterSpectra[j,i,:])*10.0 )
                        diff = subspectra - currSpectra
                        ax.plot(freq/1e6, diff)
                        
                ax.set_title('Stand: %i (%i); Dig: %i [%i]' % (antennas[i].stand.id, antennas[i].pol, antennas[i].digitizer, antennas[i].getStatus()))
                ax.set_xlabel('Frequency [MHz]')
                ax.set_ylabel('P.S.D. [dB/RBW]')
                ax.set_xlim([10,90])
                ax.set_ylim([10,80])
                
            # Save spectra image if requested
            if args.output is not None:
                base, ext = os.path.splitext(args.output)
                outFigure = "%s-%02i%s" % (base, f+1, ext)
                fig.savefig(outFigure)
                
        plt.draw()
        
    print "RBW: %.1f Hz" % (freq[1]-freq[0])
    plt.show()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
            description='read in a TBW file and create a collection of time-averaged spectra', 
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
            )
    parser.add_argument('filename', type=str, 
                        help='filename to process')
    parser.add_argument('-m', '--metadata', type=str, 
                        help='name of the SSMIF or metadata tarball file to use for mappings')
    wgroup = parser.add_mutually_exclusive_group(required=False)
    wgroup.add_argument('-t', '--bartlett', action='store_true', 
                        help='apply a Bartlett window to the data')
    wgroup.add_argument('-b', '--blackman', action='store_true', 
                        help='apply a Blackman window to the data')
    wgroup.add_argument('-n', '--hanning', action='store_true', 
                        help='apply a Hanning window to the data')
    parser.add_argument('-q', '--quiet', dest='verbose', action='store_false',
                        help='run %(prog)s in silent mode')
    parser.add_argument('-l', '--fft-length', type=aph.positive_int, default=4096, 
                        help='set FFT length')
    parser.add_argument('-g', '--gain-correct', action='store_true', 
                        help='correct signals for the cable losses')
    parser.add_argument('-s', '--stack', action='store_true', 
                        help="stack spectra in groups of 6 (if '-g' is enabled only)")
    parser.add_argument('-o', '--output', type=str, 
                        help='output file name for spectra image')
    args = parser.parse_args()
    main(args)




# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import division
import sys
if sys.version_info > (3,):
    xrange = range
    
"""
Module for calculating dispersion delay due to an ionized ISM and performing
incoherent/coherent dedispersion.

.. versionchanged:: 1.0.1
    Added support for using PyFFTW instead of NumPy for the FFTs and iFFTs.
"""

import os
import numpy

try:
    import pickle
    import pyfftw
    
    from lsl.common.paths import data as dataPath
    
    # Enable the PyFFTW cache
    if not pyfftw.interfaces.cache.is_enabled():
        pyfftw.interfaces.cache.enable()
        pyfftw.interfaces.cache.set_keepalive_time(60)
        
    # Read in the wisdom (if it exists)
    wisdomFilename = os.path.join(dataPath, 'pyfftw-wisdom.pkl')
    if os.path.exists(wisdomFilename):
        fh = open(wisdomFilename, 'r')
        wisdom = pickle.load(fh)
        fh.close()
        
        pyfftw.import_wisdom(wisdom)
        useWisdom = True
    else:
        useWisdom = False
        
    usePyFFTW = True
    
except ImportError:
    usePyFFTW = False
    useWisdom = False


__version__ = '0.5'
__revision__ = '$Rev: 2469 $'
__all__ = ['delay', 'incoherent', 'getCoherentSampleSize', 'coherent', 
        '__version__', '__revision__', '__all__']


# Dispersion constant in MHz^2 s / pc cm^-3
_D = 4.148808e3


# Coherent dedispersion N and chirp cache
_coherentCache = {}


def delay(freq, dm):
    """
    Calculate the relative delay due to dispersion over a given frequency
    range in Hz for a particular dispersion measure in pc cm^-3.  Return 
    the dispersive delay in seconds.
    
    .. versionchanged:: 1.1.1
        If only a single frequency is provided, the returned delay is 
        relative to infinite frequency.
    """
    
    # Validate in input frequencies
    ## Right Type?
    try:
        freq.size
    except AttributeError:
        freq = numpy.array(freq, ndmin=1)
    ## Right size?
    singleFreq = False
    if freq.size == 1:
        singleFreq = True
        freq = numpy.append(freq, numpy.inf)
        
    # Delay in s
    tDelay = dm*_D*((1e6/freq)**2 - (1e6/freq.max())**2)
    
    # Cleanup
    if singleFreq:
        tDelay = tDelay[0]
        
    return tDelay


def incoherent(freq, waterfall, tInt, dm, boundary='wrap', fill_value=numpy.nan):
    """
    Given a list of frequencies in Hz, a 2-D array of spectra as a function of
    time (time by frequency), and an integration time in seconds, perform 
    incoherent dedispersion on the data.
    
    The 'boundary' keyword is used to control how the boundary is treated.  The
    two options are:
      * wrap - Wrap the time boundary for each frequency (default)
      * fill - Fill the data not within the dedispersed time range with
               the value specified by the 'fill_value' keyword
    
    .. versionchanged:: 1.0.3
        Added in options for handling the boundary conditions
    """
    
    # Validate the boundary mode
    if boundary not in ('wrap', 'fill'):
        raise ValueError("Unknown boundary handling type '%s'" % boundary)
        
    # Compute the dispersive delay for the given frequency range
    tDelay = delay(freq, dm)
    
    # Convert the delays to integration periods
    tDelay = numpy.round(tDelay / tInt)
    tDelay = tDelay.astype(numpy.int32)
    
    # Roll the various frequency bins by the right amount
    ddWaterfall = waterfall*0.0
    for i,d in enumerate(tDelay):
        ddWaterfall[:,i] = numpy.roll(waterfall[:,i], -d)
        if boundary == 'fill' and d > 0:
            ddWaterfall[-d:,i] = fill_value
            
    # Return
    return ddWaterfall


def getCoherentSampleSize(centralFreq, sampleRate, dm):
    """
    Estimate the number of samples needed to successfully apply coherent 
    dedispersion to a data stream.
    """
    
    # Roughly estimate the number of points we need to look at to do the dedispersion 
    # correctly.  Based on the the relative dispersion delay between the high and low
    # ends of an observational band.
    F0 = centralFreq
    BW = sampleRate

    delayBand = dm*_D *((1e6/(F0-BW/2.0))**2 - (1e6/(F0+BW/2.0))**2)	# Dispersion delay across the band
    samples = delayBand*BW									# Conversion to samples
    samples = 2**(numpy.ceil(numpy.log(samples)/numpy.log(2)))		# Conversion to next largest power of 2
    samples *= 2											# Correction for the 'wings' of the convolution
    
    return int(samples)


def __taperFunction(freq):
    """
    Taper function based Equation (1) of "Pulsar Coherent De-dispersion 
    Experiment at Urumqi Observatory" CJA&A, 2006, S2, 53.
    """

    freqMHz = freq / 1e6
    fMHz0 = freqMHz.mean()
    fMHz1 = freqMHz - fMHz0
    BW = fMHz1.max() - fMHz1.min()

    taper = 1.0 / numpy.sqrt( 1.0 + ( numpy.abs(fMHz1) / (0.47*BW) )**80 )

    return taper


def __chirpFunction(freq, dm, taper=False):
    """
    Chip function for coherent dedispersion for a given set of frequencies (in Hz).  
    Based on Equation (6) of "Pulsar Observations II -- Coherent Dedispersion, 
    Polarimetry, and Timing" By Stairs, I. H.
    """
    
    freqMHz = freq / 1e6
    fMHz0 = freqMHz.mean()
    fMHz1 = freqMHz - fMHz0
    
    chirp = numpy.exp(-2j*numpy.pi*_D*1e6 / (fMHz0**2*(fMHz0 + fMHz1)) * dm*fMHz1**2)
    if taper:
        chirp *= __taperFunction(freq)
    
    return chirp


def coherent(t, timeseries, centralFreq, sampleRate, dm, taper=False, previousTime=None, previousData=None, nextTime=None, nextData=None, enableCaching=True):
    """
    Simple coherent dedispersion of complex-valued time-series data at a given central
    frequency and sample rate.  A tapering function can also be applied to the chirp of 
    the form:
        
        :math:`\\sqrt{1 + \\left(\\frac{\\Delta f_{MHz}}{0.47 \\times \\mbox{BW}}\\right)^{80}}`, 
        
    where :math:`\\Delta f_{MHz}` is the frequency difference in MHz from the band 
    center and BW is the bandwidth in MHz.
    
    .. note::
        At the large fractional bandwidths of LWA, the window size needed for coherent 
        dedispersion can be prohibitive.  For example, at 74 MHz with 19.6 MS/s and a
        DM or 10 pc / cm^3 this function uses a window size of about 268 million points.
        
    .. versionchanged:: 1.0.1
        Added support for using PyFFTW instead of NumPy for the FFTs and iFFTs.
        Added a cache for storing the chrip function between subsequent calls
        
    .. versionchanged:: 0.6.4
        Added support for keeping track of time through the dedispersion process.
    """
    
    # Caching for N and the chirp function
    try:
        pair = (centralFreq*1.0, sampleRate*1.0, dm*1.0, taper, timeseries.dtype)
        
        # Get an idea of how many samples we need to do the dedispersion correctly
        # Compute the chirp function 
        N, chirp = _coherentCache[pair]
    except KeyError:
        # Get an idea of how many samples we need to do the dedispersion correctly
        N = getCoherentSampleSize(centralFreq, sampleRate, dm)
        
        # Compute the chirp function 
        freq = numpy.fft.fftfreq(N, d=1.0/sampleRate) + centralFreq 
        chirp = __chirpFunction(freq, dm, taper=taper) 
        chirp = chirp.astype(timeseries.dtype)
        
        # Update the cache
        if enableCaching:
            _coherentCache[pair] = (N, chirp)
            
    # Figure out the output array size
    nSets = len(timeseries) // N
    outT = numpy.zeros(timeseries.size, dtype=t.dtype)
    outD = numpy.zeros(timeseries.size, dtype=timeseries.dtype)
    
    if nSets == 0:
        RuntimeWarning("Too few data samples for proper dedispersion")
        
    if usePyFFTW:
        if timeseries.dtype not in (numpy.complex64, numpy.complex128):
            raise RuntimeError("Unsupported data type for timeseries: %s" % str(timeseries.dtype))
            
        dd = timeseries.dtype
        di1 = numpy.empty(N, dtype=dd)
        do1 = numpy.empty(N, dtype=dd)
        do2 = numpy.empty(N, dtype=dd)
            
        forwardPlan = pyfftw.FFTW(di1, do1, direction='FFTW_FORWARD', flags=('FFTW_ESTIMATE', 'FFTW_UNALIGNED'))
        backwardPlan = pyfftw.FFTW(do1, do2, direction='FFTW_BACKWARD', flags=('FFTW_ESTIMATE', 'FFTW_UNALIGNED'))
        
    # Go!
    for i in xrange(2*nSets+1):
        start = i*N/2 - N/4
        stop = start + N
        
        if start < 0:
            timeIn = numpy.zeros(N, dtype=t.dtype)
            dataIn = numpy.zeros(N, dtype=timeseries.dtype)
            
            if previousData is not None:
                try:
                    timeIn[:-start] = previousTime[start:]
                    dataIn[:-start] = previousData[start:]
                except ValueError:
                    raise RuntimeError("Too few data samples for proper start buffering")
                    
            timeIn[-start:N] = t[0:N+start]
            dataIn[-start:N] = timeseries[0:N+start]
            
        elif stop > timeseries.size:
            timeIn = numpy.zeros(N, dtype=t.dtype)
            dataIn = numpy.zeros(N, dtype=timeseries.dtype)
            
            if nextData is not None:
                try:
                    timeIn[timeseries.size-start:] = nextTime[:(dataIn.size-(timeseries.size-start))]
                    dataIn[timeseries.size-start:] = nextData[:(dataIn.size-(timeseries.size-start))]
                except ValueError:
                    raise RuntimeError("Too few data samples for proper end buffering")
                    
            if start < timeseries.size:
                timeIn[0:(timeseries.size-start)] = t[start:]
                dataIn[0:(timeseries.size-start)] = timeseries[start:]
                
        else:
            timeIn = t[start:stop]
            dataIn = timeseries[start:stop]
            
        timeOut = timeIn
        if usePyFFTW:
            forwardPlan(dataIn, do1)
            do1 *= chirp
            backwardPlan(do1, do2)
            dataOut = do2
            
        else:
            dataOut = numpy.fft.fft( dataIn )
            dataOut *= chirp
            dataOut = numpy.fft.ifft( dataOut )
            
        # Get the output data ranges
        outStart  = i*N//2
        outStop   = outStart + N//2
        dataStart = N//4
        dataStop  = dataStart + N//2
        
        # Make sure we don't fall off the end of the array
        if outStop >= outD.size:
            diff = outStop - outD.size
            outStop  -= diff
            dataStop -= diff
            
        if i == 0 and previousData is None:
            continue
        if i == (2*nSets) and nextData is None:
            continue
            
        outT[outStart:outStop] = timeOut[dataStart:dataStop]
        outD[outStart:outStop] = dataOut[dataStart:dataStop]
        
    return outT, outD



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function

"""
Modules to take TBW/TBN time series data and write it to a FITS file composed 
of binary tables.  The format of TSFITS files is:

**PRIMARY HDU**

Keywords:
 * OBJECT - object being observed in this data set, e.g., 'zenith'
 * TELESCOP - telescope used for the observations, e.g., 'LWA-1'
 * OBSMODE - observation mode used for the data.  Options are TBW or TBN.
 * NSTANDS - number of stands found in the file
 * TBWBITS *optional* - how many bits are used by the TBW data (4 or 12)
 * FILTER *optional* - TBN filter code used to acquire the data
 * GAIN *optional* - TBN gain setting used to acquire the data
 * FREQ *optional* - central frequency in Hz for the TBN observations

**TIME SERIES HDU**

Binary table that stores the actual observations.  There is one TIME SERIES 
extension per stand (TBW/TBN) or beam/tuning pair (DRX) in the data.

Keywords:
 * EXTNAME - extension name of 'TIME SERIES'
 * EXTVER - extension version equal to the order in which the data was 
            added to the FITS file
 * STAND - stand number associated with this data
 * DATE-OBS - date of observation for the first sample of the first data 
              row

Data:
 * DATA - column storing the observations.  For TBW data this consists of 
          400 (12-bit) or 1200 (4-bit) elements per row stored as 16-bit 
          integers.  For TBN data this consists of 512 elements per row 
          stored as 32-bit complex numbers.  For DRX data this consists of 
          4096 elements per row stored as 32-bit complex numbers.
 * POL - column storing the polarization associated with the DATA column.  
         'x' polarization is stored as 0 and 'y' polarization is stored as 1.  
         The values is stored as a 16-bit integer.
 * TIME - number of samples at f_S (196 MSamples/s) since DATE-OBS for the
          first sample in the row.  This is stored as a 64-bit integer.
"""

import os
import time
import numpy
import pyfits
from datetime import datetime, timedelta, tzinfo

from lsl.common import dp as dp_common

__version__ = '0.5'
__revision__ = '$Rev: 2667 $'
__all__ = ['UTC', 'TSFITS', 'TBW', 'TBN', '__version__', '__revision__', '__all__']


import warnings
warnings.warn("lsl.writer.tsfits will be removed in LSL version 1.3.0", DeprecationWarning)


class UTC(tzinfo):
    """
    tzinfo object for UTC time.
    """

    def utcoffset(self, dt):
        return timedelta(0)

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return timedelta(0)


class TSFITS(object):
    """
    Class that holds TSFITS data until it is ready to be written to disk.
    """

    def __init__(self, filename, mode, UseQueue=True, verbose=False, memmap=None, clobber=False):
        """
        Initialize a TSFITS object using a filename and an observation mode 
        (TBW or TBN).  Optionally, TSFITS can be told to overwrite the file if it 
        already exists using the 'Overwrite' keyword.
        
        .. versionchanged:: 1.1.2
            Added the 'memmap' to control if the file is memory mapped.
            Changed 'Overwrite' to 'clobber' for consistency with other writers.
        """

        assert(mode in ['TBW', 'TBN'])

        self.filename = filename
        self.mode = mode
        self.hdulist = None
        self.UseQueue = UseQueue
        self.queue = {}
        self.queueLimit = 10000
        self.site = 'LWA'
        self.firstSamples = {}
        self.verbose = verbose

        if os.path.exists(filename):
            if clobber:
                os.unlink(filename)
            else:
                raise IOError("File '%s' already exists" % filename)
        self.standCount = 0
        primary = pyfits.PrimaryHDU()
        primary.header['OBJECT'] = 'zenith'
        primary.header['TELESCOP'] = self.site
        primary.header['OBSMODE'] = self.mode
        primary.header['NSTAND'] = self.standCount
        
        primary.header['COMMENT'] = " FITS (Flexible Image Transport System) format is defined in 'Astronomy and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H"
        
        hdulist = pyfits.HDUList([primary])
        hdulist.writeto(filename)
        
        self.hdulist = pyfits.open(self.filename, mode="update", memmap=memmap)

    def info(self):
        """
        Short-cut to the pyfits.info() function on an opened FITS file.
        """

        self.hdulist.info()

    def flush(self):
        """
        Short-cut to the pyfits.flush() function on an opened FITS file.
        """

        self.hdulist.flush()

    def close(self):
        """
        Empty the data queue (if it exists) and write all changes to disk 
        using flush.
        """

        if self.UseQueue:
            self.__emptyQueue()
        self.flush()

    def setSite(self, site):
        """
        Set the TELESCOP keyword in the primary HDU using an lsl.common.stations
        object.
        """

        self.site = site.name

        self.hdulist[0].header['TELESCOP'] = self.site

        self.flush()

    def __findExtension(self, stand):
        """
        Private function to find out which extension stores the stand in 
        question.  None is returned is that stand is not in the TSFITS file.
        """

        extension = None

        i = 1
        for hdu in self.hdulist[1:]:
            if int(hdu.header['stand']) == int(stand):
                extension = i
                break
            i = i + 1

        return extension

    def __makeAppendTable(self, extension, AddRows=1):
        """
        Private function to make a temporary table for appending data.
        """

        nrows = self.hdulist[extension].data.shape[0]
        tempHDU = pyfits.new_table(self.hdulist[extension].columns, nrows=nrows+AddRows)
        tempHDU.header = self.hdulist[extension].header.copy()
    
        return tempHDU

    def __applyAppendTable(self, extension, tempHDU):
        """
        Private function to replace the given extension with the temporary
        table.
        """

        self.hdulist[extension] = tempHDU
        self.flush()

    def __addDataSingle(self, frame):
        """
        Private function to add a single data entry to a TSFITS file.  This 
        method is not particular fast since the existing file needs to be copied
        to append data to the end of a particular extension.
        """

        if self.mode == 'TBW':
            stand = frame.parseID()
        else:
            try:
                stand, pol = frame.parseID()
            except ValueError:
                stand, tune, pol = frame.parseID()
                stand = stand*10 + tune
        extension = self.__findExtension(stand)

        if extension is None:
            if self.verbose:
                print("Stand '%i' not found, creating new binary table extension" % stand)
            self.standCount = self.standCount + 1

            if self.mode == 'TBW':
                # Data
                c1 = pyfits.Column(name='data', format='%iI' % frame.data.xy.shape[1], array=frame.data.xy.astype(numpy.int16))
                # Polarization
                c2 = pyfits.Column(name='pol', format='1I', array=numpy.array([0, 1], dtype=numpy.int16))
                # Time
                c3 = pyfits.Column(name='time', format='1K', array=numpy.array([0, 0]))
            else:
                # Data
                data = frame.data.iq.astype(numpy.csingle)
                data.shape = (1,)+data.shape
                c1 = pyfits.Column(name='data', format='%iC' % data.shape[1], array=data)
                # Polarization
                c2 = pyfits.Column(name='pol', format='1I', array=numpy.array([0], dtype=numpy.int16))
                # Time
                c3 = pyfits.Column(name='time', format='1K', array=numpy.array([0]))

            # Define the collection of columns
            colDefs = pyfits.ColDefs([c1, c2, c3])

            # Get the time of the first sample and convert it to a datetime
            self.firstSamples[stand] = long(frame.data.timeTag / dp_common.fS)
            firstSample = datetime.utcfromtimestamp(self.firstSamples[stand])

            tsfits = pyfits.new_table(colDefs)
            tsfits.header.set('EXTNAME', 'TIME SERIES', after='TFIELDS')
            tsfits.header.set('EXTVER', self.standCount, after='EXTNAME')
            tsfits.header.set('STAND', stand, after='EXTVER')
            tsfits.header['DATE-OBS'] = firstSample.isoformat()

            self.hdulist.append(tsfits)
            self.hdulist[0].header['NSTAND'] = self.standCount
            self.flush()
            
            self.hdulist[-1].data.field('pol')[0] = numpy.array([0], dtype=numpy.int16)
            self.hdulist[-1].data.field('time')[0] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
            if self.mode == 'TBW':
                self.hdulist[-1].data.field('pol')[1] = numpy.array([1], dtype=numpy.int16)
                self.hdulist[-1].data.field('time')[1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
        else:
            # Make sure that we have a first sample time to reference to if 
            # we are adding on to the end of a file
            if stand not in self.firstSamples.keys():
                firstSample = datetime.strptime(self.hdulist[extension].header['DATE-OBS'], "%Y-%m-%dT%H:%M:%S")
                self.firstSamples[stand] = long(time.mktime(firstSample.timetuple()))

            nrows = self.hdulist[extension].data.shape[0]
            if self.mode == 'TBW':
                tempHDU = self.__makeAppendTable(extension, AddRows=2)
                tempHDU.data.field('data')[nrows] = numpy.squeeze(frame.data.xy.astype(numpy.int16)[0,:])
                tempHDU.data.field('pol')[nrows] = numpy.array([0], dtype=numpy.int16)
                tempHDU.data.field('time')[nrows] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
                tempHDU.data.field('data')[nrows+1] = numpy.squeeze(frame.data.xy.astype(numpy.int16)[1,:])
                tempHDU.data.field('pol')[nrows+1] = numpy.array([1], dtype=numpy.int16)
                tempHDU.data.field('time')[nrows+1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
            else:
                tempHDU = self.__makeAppendTable(extension, AddRows=1)
                data = frame.data.iq.astype(numpy.csingle)
                data.shape = (1,)+data.shape
                tempHDU.data.field('data')[nrows:] = data
                tempHDU.data.field('pol')[nrows:] = numpy.array([pol], dtype=numpy.int16)
                tempHDU.data.field('time')[nrows:] = numpy.array([frame.data.timeTag]) / dp_common.fS - self.firstSamples[stand]

            self.__applyAppendTable(extension, tempHDU)

        self.flush()

    def __addDataQueue(self, frame):
        """
        Private function similar to __addDataSignle, but it saves the data to 
        memory (self.queue) to be written once the queue fills up.
        """

        if self.mode == 'TBW':
            stand = frame.parseID()
        else:
            try:
                stand, pol = frame.parseID()
            except ValueError:
                stand, tune, pol = frame.parseID()
                stand = stand*10 + tune
        if stand not in self.queue.keys():
            self.queue[stand] = []
        self.queue[stand].append(frame)

        if len(self.queue[stand]) >= self.queueLimit:
            start = 0
            extension = self.__findExtension(stand)

            if extension is None:
                start = 1
                frame = self.queue[stand][0]
                
                if self.verbose:
                    print("Stand '%i' not found, creating new binary table extension" % stand)
                self.standCount = self.standCount + 1

                if self.mode == 'TBW':
                    # Data
                    c1 = pyfits.Column(name='data', format='%iI' % frame.data.xy.shape[1], array=frame.data.xy.astype(numpy.int16))
                    # Polarization
                    c2 = pyfits.Column(name='pol', format='1I', array=numpy.array([0, 1], dtype=numpy.int16))
                    # Time
                    c3 = pyfits.Column(name='time', format='1K', array=numpy.array([0, 0]))
                else:
                    # Data
                    data = frame.data.iq.astype(numpy.csingle)
                    data.shape = (1,)+data.shape
                    c1 = pyfits.Column(name='data', format='%iC' % data.shape[1], array=data)
                    # Polarization
                    c2 = pyfits.Column(name='pol', format='1I', array=numpy.array([0], dtype=numpy.int16))
                    # Time
                    c3 = pyfits.Column(name='time', format='1K', array=numpy.array([0]))

                # Define the collection of columns
                colDefs = pyfits.ColDefs([c1, c2, c3])

                # Get the time of the first sample and convert it to a datetime
                self.firstSamples[stand] = long(frame.data.timeTag / dp_common.fS)
                firstSample = datetime.utcfromtimestamp(self.firstSamples[stand])

                tsfits = pyfits.new_table(colDefs)
                tsfits.header.set('EXTNAME', 'TIME SERIES', after='TFIELDS')
                tsfits.header.set('EXTVER', self.standCount, after='EXTNAME')
                tsfits.header.set('STAND', stand, after='EXTVER')
                tsfits.header['DATE-OBS'] = firstSample.isoformat('T')

                self.hdulist.append(tsfits)
                self.hdulist[0].header['NSTAND'] = self.standCount
                self.flush()
                
                self.hdulist[-1].data.field('pol')[0] = numpy.array([0], dtype=numpy.int16)
                self.hdulist[-1].data.field('time')[0] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
                if self.mode == 'TBW':
                    self.hdulist[-1].data.field('time')[1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
                    self.hdulist[-1].data.field('pol')[1] = numpy.array([1], dtype=numpy.int16)

                self.flush()
                extension = self.__findExtension(stand)

            # Make sure that we have a first sample time to reference to if 
            # we are adding on to the end of a file
            if stand not in self.firstSamples.keys():
                firstSample = datetime.strptime(self.hdulist[extension].header['DATE-OBS'], "%Y-%m-%dT%H:%M:%S")
                self.firstSamples[stand] = long(time.mktime(firstSample.timetuple()))

            nrows = self.hdulist[extension].data.shape[0]
            if self.mode == 'TBW':
                tempHDU = self.__makeAppendTable(extension, AddRows=2*(self.queueLimit-start))
                for count,frame in zip(range(len(self.queue[stand][start:])), self.queue[stand][start:]):
                    tempHDU.data.field('data')[nrows+2*count] = numpy.squeeze(frame.data.xy.astype(numpy.int16)[0,:])
                    tempHDU.data.field('pol')[nrows+2*count] = numpy.array([0], dtype=numpy.int16)
                    tempHDU.data.field('time')[nrows+2*count] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
                    tempHDU.data.field('data')[nrows+2*count+1] = numpy.squeeze(frame.data.xy.astype(numpy.int16)[1,:])
                    tempHDU.data.field('pol')[nrows+2*count+1] = numpy.array([1], dtype=numpy.int16)
                    tempHDU.data.field('time')[nrows+2*count+1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
            else:
                tempHDU = self.__makeAppendTable(extension, AddRows=(self.queueLimit-start))
                for count,frame in zip(range(len(self.queue[stand][start:])), self.queue[stand][start:]):
                    try:
                        stand, pol = frame.parseID()
                    except ValueError:
                        stand, tune, pol = frame.parseID()
                        stand = stand*10 + tune
                        
                    data = frame.data.iq.astype(numpy.csingle)
                    data.shape = (1,) + data.shape
                    tempHDU.data.field('data')[nrows+count] = data
                    tempHDU.data.field('pol')[nrows+count] = numpy.array([pol], dtype=numpy.int16)
                    tempHDU.data.field('time')[nrows+count] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]

            self.__applyAppendTable(extension, tempHDU)
            del(self.queue[stand])

    def __emptyQueue(self):
        """
        Private function to empty a empty the self.queue dictionary on demand
        if it exists.
        """

        for stand in self.queue.keys():
            if len(self.queue[stand]) == 0:
                continue
            queueSize = len(self.queue[stand])
            start = 0
            extension = self.__findExtension(stand)

            if extension is None:
                start = 1
                frame = self.queue[stand][0]
                
                if self.verbose:
                    print("Stand '%i' not found, creating new binary table extension" % stand)
                self.standCount = self.standCount + 1

                if self.mode == 'TBW':
                    # Data
                    c1 = pyfits.Column(name='data', format='%iI' % frame.data.xy.shape[1], array=frame.data.xy.astype(numpy.int16))
                    # Polarization
                    c2 = pyfits.Column(name='pol', format='1I', array=numpy.array([0, 1], dtype=numpy.int16))
                    # Time
                    c3 = pyfits.Column(name='time', format='1K', array=numpy.array([0, 0]))
                else:
                    # Data
                    data = frame.data.iq.astype(numpy.csingle)
                    data.shape = (1,)+data.shape
                    c1 = pyfits.Column(name='data', format='%iC' % data.shape[1], array=data)
                    # Polarization
                    c2 = pyfits.Column(name='pol', format='1I', array=numpy.array([0], dtype=numpy.int16))
                    # Time
                    c3 = pyfits.Column(name='time', format='1K', array=numpy.array([0]))

                # Define the collection of columns
                colDefs = pyfits.ColDefs([c1, c2, c3])

                # Get the time of the first sample and convert it to a datetime
                self.firstSamples[stand] = long(frame.data.timeTag / dp_common.fS)
                firstSample = datetime.utcfromtimestamp(self.firstSamples[stand])

                tsfits = pyfits.new_table(colDefs)
                tsfits.header.set('EXTNAME', 'TIME SERIES', after='TFIELDS')
                tsfits.header.set('EXTVER', self.standCount, after='EXTNAME')
                tsfits.header.set('STAND', stand, after='EXTVER')
                tsfits.header['DATE-OBS'] = firstSample.isoformat('T')

                self.hdulist.append(tsfits)
                self.hdulist[0].header['NSTAND'] = self.standCount
                self.flush()
                
                self.hdulist[-1].data.field('pol')[0] = numpy.array([0], dtype=numpy.int16)
                self.hdulist[-1].data.field('time')[0] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
                if self.mode == 'TBW':
                    self.hdulist[-1].data.field('time')[1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
                    self.hdulist[-1].data.field('pol')[1] = numpy.array([1], dtype=numpy.int16)

                self.flush()
                extension = self.__findExtension(stand)

            # Make sure that we have a first sample time to reference to if 
            # we are adding on to the end of a file
            if stand not in self.firstSamples.keys():
                firstSample = datetime.strptime(self.hdulist[extension].header['DATE-OBS'], "%Y-%m-%dT%H:%M:%S")
                self.firstSamples[stand] = long(time.mktime(firstSample.timetuple()))

            nrows = self.hdulist[extension].data.shape[0]
            if self.mode == 'TBW':
                tempHDU = self.__makeAppendTable(extension, AddRows=2*(queueSize-start))
                for count,frame in zip(range(len(self.queue[stand][start:])), self.queue[stand][start:]):
                    tempHDU.data.field('data')[nrows+2*count] = numpy.squeeze(frame.data.xy.astype(numpy.int16)[0,:])
                    tempHDU.data.field('pol')[nrows+2*count] = numpy.array([0], dtype=numpy.int16)
                    tempHDU.data.field('time')[nrows+2*count] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
                    tempHDU.data.field('data')[nrows+2*count+1] = numpy.squeeze(frame.data.xy.astype(numpy.int16)[1,:])
                    tempHDU.data.field('pol')[nrows+2*count+1] = numpy.array([1], dtype=numpy.int16)
                    tempHDU.data.field('time')[nrows+2*count+1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
            else:
                tempHDU = self.__makeAppendTable(extension, AddRows=(queueSize-start))
                for count,frame in zip(range(len(self.queue[stand][start:])), self.queue[stand][start:]):
                    try:
                        stand, pol = frame.parseID()
                    except ValueError:
                        stand, tune, pol = frame.parseID()
                        stand = stand*10 + tune
                        
                    data = frame.data.iq.astype(numpy.csingle)
                    data.shape = (1,) + data.shape
                    tempHDU.data.field('data')[nrows+count] = data
                    tempHDU.data.field('pol')[nrows+count] = numpy.array([pol], dtype=numpy.int16)
                    tempHDU.data.field('time')[nrows+count] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]

            self.__applyAppendTable(extension, tempHDU)
            del(self.queue[stand])

    def addStandData(self, frame):
        """
        Add a frame object to the TSFITS file.  This function takes care of 
        figuring out which extension the data goes to and how it should be formated.
        This function also updates the primary HDU with information about the data, 
        i.e., TBW data bits, TBN filter code, etc.
        """

        if self.mode == 'TBW':
            try:
                self.hdulist[0].header['TBWBITS']
            except:
                self.hdulist[0].header['TBWBITS'] = frame.getDataBits()
                self.flush()
        else:
            try:
                self.hdulist[0].header['FILTER']
            except:
                if frame.getFilterCode() is not None:
                    self.hdulist[0].header['FILTER'] = frame.getFilterCode()
                if getattr(frame.header, 'gain', None) is not None:
                    self.hdulist[0].header['GAIN'] = frame.header.gain
                if getattr(frame.header, 'tuningWord', None) is not None:
                    self.hdulist[0].header['FREQ'] = frame.getCentralFreq()
                self.flush()

        if self.UseQueue:
            self.__addDataQueue(frame)
        else:
            self.__addDataSingle(frame)

    def getStandData(self, stand):
        """
        Retrieve a dictionary of all data stored for a particular stand.  The 
        dictionary keys are:
         * *data* - numpy array of data
         * *pol* - numpy array of polarizations
         * *time* - numpy array of times in samples at f_S since DATE-OBS
        """

        extension = self.__findExtension(stand)
        
        if extension is None:
            output = {}
            for key in ['data', 'pol', 'time']:
                output[key] = None
        else:
            output = {}
            for key in ['data', 'pol', 'time']:
                output[key] = self.hdulist[extension].data.field(key)
        return output


class TBW(TSFITS):
    """
    Sub-class of TSFITS for dealing with TBW data in particular.
    """

    def __init__(self, filename, UseQueue=True, verbose=False, memmap=None, clobber=False):
        super(TBW, self).__init__(filename, 'TBW', UseQueue=UseQueue, verbose=verbose, memmap=memmap, clobber=clobber)
        

class TBN(TSFITS):
    """
    Sub-class of TSFITS for dealing with TBN data in particular.
    """

    def __init__(self, filename, UseQueue=True, verbose=False, memmap=None, clobber=False):
        super(TBN, self).__init__(filename, 'TBN', UseQueue=UseQueue, verbose=verbose, memmap=memmap, clobber=clobber)



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import division
import sys
if sys.version_info > (3,):
    xrange = range
    
"""
Unit test for the lsl.reader.buffer module.
"""

import os
import unittest

from lsl.common.paths import dataBuild as dataPath
from lsl.reader import tbn, drx, tbf
from lsl.reader import errors
from lsl.reader import buffer


__revision__ = "$Rev: 2470 $"
__version__  = "0.2"
__author__    = "Jayce Dowell"


tbnFile = os.path.join(dataPath, 'tests', 'tbn-test.dat')
drxFile = os.path.join(dataPath, 'tests', 'drx-test.dat')
tbfFile = os.path.join(dataPath, 'tests', 'tbf-test.dat')


class buffer_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.reader.buffer
    module."""
    
    #
    # TBN
    #
    
    def test_tbn_default(self):
        """Test the TBN ring buffer with the default values."""
        
        fh = open(tbnFile, 'rb')
        nFpO = tbn.getFramesPerObs(fh)
        nFpO = nFpO[0] + nFpO[1]
        
        # Create the FrameBuffer instance
        frameBuffer = buffer.TBNFrameBuffer(stands=range(1,nFpO//2+1), pols=[0, 1])
        
        # Go
        while True:
            try:
                cFrame = tbn.readFrame(fh)
            except errors.eofError:
                break
            except errors.syncError:
                continue
                
            frameBuffer.append(cFrame)
            cFrames = frameBuffer.get()
            
            if cFrames is None:
                continue
            
        fh.close()
        
        # Make sure we have the right number of frames in the buffer
        nFrames = 0
        for key in frameBuffer.buffer.keys():
            nFrames = nFrames + len(frameBuffer.buffer[key])
        self.assertEqual(nFrames, 29)
        
        # Make sure nothing has happened that shouldn't have
        self.assertEqual(frameBuffer.full,    0)
        self.assertEqual(frameBuffer.partial, 0)
        self.assertEqual(frameBuffer.missing, 0)
        self.assertEqual(frameBuffer.dropped, 0)
        
        # Make sure we have the right keys
        for key in frameBuffer.buffer.keys():
            self.assertTrue(key in (119196674956800, 119196675960320))
            
        # Make sure the buffer keys have the right sizes
        self.assertEqual(len(frameBuffer.buffer[119196674956800]), 20)
        self.assertEqual(len(frameBuffer.buffer[119196675960320]),  9)
        
    def test_tbn_small(self):
        """Test a small version of the TBN ring buffer."""
        
        fh = open(tbnFile, 'rb')
        nFpO = tbn.getFramesPerObs(fh)
        nFpO = nFpO[0] + nFpO[1]
        
        # Create the FrameBuffer instance
        frameBuffer = buffer.TBNFrameBuffer(stands=range(1,nFpO//2+1), pols=[0, 1], nSegments=1)
        
        # Go
        while True:
            try:
                cFrame = tbn.readFrame(fh)
            except errors.eofError:
                break
            except errors.syncError:
                continue
                
            frameBuffer.append(cFrame)
            cFrames = frameBuffer.get()
            
            if cFrames is None:
                continue
                
            # Make sure the dump has one of the expected time tags
            self.assertTrue(cFrames[0].data.timeTag in (119196674956800, 119196675960320))
            
            # Make sure it has the right number of frames
            self.assertEqual(len(cFrames), nFpO)
            
        fh.close()
        
    def test_tbn_buffer_reorder(self):
        """Test the reorder function of the TBN ring buffer."""
        
        fh = open(tbnFile, 'rb')
        nFpO = tbn.getFramesPerObs(fh)
        nFpO = nFpO[0] + nFpO[1]
        
        # Create the FrameBuffer instance
        frameBuffer = buffer.TBNFrameBuffer(stands=range(1,nFpO//2+1), pols=[0, 1], nSegments=1, ReorderFrames=True)
        
        # Go
        while True:
            try:
                cFrame = tbn.readFrame(fh)
            except errors.eofError:
                break
            except errors.syncError:
                continue
                
            frameBuffer.append(cFrame)
            cFrames = frameBuffer.get()
            
            if cFrames is None:
                continue
                
            # Make sure the dump has one of the expected time tags
            self.assertTrue(cFrames[0].data.timeTag in (119196674956800, 119196675960320))
            
            # Make sure it has the right number of frames
            self.assertEqual(len(cFrames), nFpO)
            
            # Check the order
            for i in xrange(1, len(cFrames)):
                pS, pP = cFrames[i-1].parseID()
                cS, cP = cFrames[i].parseID()
                
                pID = pS*2 + pP
                cID = cS*2 + cP
                self.assertTrue(cID > pID)
                
        fh.close()
        
    def test_tbn_buffer_flush(self):
        """Test the TBN ring buffer's flush() function."""
        
        fh = open(tbnFile, 'rb')
        nFpO = tbn.getFramesPerObs(fh)
        nFpO = nFpO[0] + nFpO[1]
        
        # Create the FrameBuffer instance
        frameBuffer = buffer.TBNFrameBuffer(stands=range(1,nFpO//2+1), pols=[0, 1])
        
        # Go
        while True:
            try:
                cFrame = tbn.readFrame(fh)
            except errors.eofError:
                break
            except errors.syncError:
                continue
                
            frameBuffer.append(cFrame)
            cFrames = frameBuffer.get()
            
            if cFrames is None:
                continue
                
        fh.close()
        
        # Flush the buffer
        for cFrames in frameBuffer.flush():
            # Make sure the dump has one of the expected time tags
            self.assertTrue(cFrames[0].data.timeTag in (119196674956800, 119196675960320))
            
            # Make sure it has the right number of frames
            self.assertEqual(len(cFrames), nFpO)
            
    def test_tbn_buffer_reorder_flush(self):
        """Test the TBN ring buffer's flush() function with reordering."""
        
        fh = open(tbnFile, 'rb')
        nFpO = tbn.getFramesPerObs(fh)
        nFpO = nFpO[0] + nFpO[1]
        
        # Create the FrameBuffer instance
        frameBuffer = buffer.TBNFrameBuffer(stands=range(1,nFpO//2+1), pols=[0, 1], ReorderFrames=True)
        
        # Go
        while True:
            try:
                cFrame = tbn.readFrame(fh)
            except errors.eofError:
                break
            except errors.syncError:
                continue
                
            frameBuffer.append(cFrame)
            cFrames = frameBuffer.get()
            
            if cFrames is None:
                continue
                
        fh.close()
        
        # Flush the buffer
        for cFrames in frameBuffer.flush():
            # Make sure the dump has one of the expected time tags
            self.assertTrue(cFrames[0].data.timeTag in (119196674956800, 119196675960320))
            
            # Make sure it has the right number of frames
            self.assertEqual(len(cFrames), nFpO)
            
            # Check the order
            for i in xrange(1, len(cFrames)):
                pS, pP = cFrames[i-1].parseID()
                cS, cP = cFrames[i].parseID()
                
                pID = pS*2 + pP
                cID = cS*2 + cP
                self.assertTrue(cID > pID)
                
    #
    # DRX
    #
    
    def test_drx_basic(self):
        """Test the DRX ring buffer."""
        
        fh = open(drxFile, 'rb')
        junkFrame = drx.readFrame(fh)
        b,t,p = junkFrame.parseID()
        fh.seek(-drx.FrameSize, 1)
        
        # Create the FrameBuffer instance
        frameBuffer = buffer.DRXFrameBuffer(beams=[b,], tunes=[1, 2], pols=[0, 1], nSegments=2)
        
        # Go
        dumped = []
        while True:
            try:
                cFrame = drx.readFrame(fh)
            except errors.eofError:
                break
            except errors.syncError:
                continue

            frameBuffer.append(cFrame)
            cFrames = frameBuffer.get()

            if cFrames is None:
                continue
                
            dumped.append( cFrames[0].data.timeTag )
            
        fh.close()
        
        # Make sure we have the right number of frames in the buffer
        nFrames = 0
        for key in frameBuffer.buffer.keys():
            nFrames = nFrames + len(frameBuffer.buffer[key])
        self.assertEqual(nFrames, 1)
        self.assertEqual(nFrames+len(dumped)*4, 32+1)
        
        # Make sure nothing has happened that shouldn't have
        self.assertEqual(frameBuffer.full,    7)
        self.assertEqual(frameBuffer.partial, 1)
        self.assertEqual(frameBuffer.missing, 1)
        self.assertEqual(frameBuffer.dropped, 0)
        
        # Make sure we have the right keys
        for key in dumped:
            self.assertTrue(key in (257355782095018376, 257355782095059336, 257355782095100296, 257355782095141256, 257355782095182216, 257355782095223176, 257355782095264136, 257355782095305096))
            
        for key in frameBuffer.buffer.keys():
            self.assertTrue(key in (257355782095346056,))
        
        # Make sure the buffer keys have the right sizes
        self.assertEqual(len(frameBuffer.buffer[257355782095346056]), 1)
        
    def test_drx_reorder(self):
        """Test the reorder function of the TBN ring buffer."""
        
        fh = open(drxFile, 'rb')
        junkFrame = drx.readFrame(fh)
        b,t,p = junkFrame.parseID()
        fh.seek(-drx.FrameSize, 1)
        
        # Create the FrameBuffer instance
        frameBuffer = buffer.DRXFrameBuffer(beams=[b,], tunes=[1, 2], pols=[0, 1], nSegments=2, ReorderFrames=True)
        
        # Go
        while True:
            try:
                cFrame = drx.readFrame(fh)
            except errors.eofError:
                break
            except errors.syncError:
                continue

            frameBuffer.append(cFrame)
            cFrames = frameBuffer.get()

            if cFrames is None:
                continue
                
            # Make sure it has the right number of frames
            self.assertEqual(len(cFrames), 4)
            
            # Check the order
            for i in xrange(1, len(cFrames)):
                pB, pT, pP = cFrames[i-1].parseID()
                cB, cT, cP = cFrames[i].parseID()
                
                pID = 4*pB + 2*(pT-1) + pP
                cID = 4*cB + 2*(cT-1) + cP
                self.assertTrue(cID > pID)
                
        fh.close()
        
    def test_drx_buffer_flush(self):
        """Test the DRX ring buffer's flush() function."""
        
        fh = open(drxFile, 'rb')
        junkFrame = drx.readFrame(fh)
        b,t,p = junkFrame.parseID()
        fh.seek(-drx.FrameSize, 1)
        
        # Create the FrameBuffer instance
        frameBuffer = buffer.DRXFrameBuffer(beams=[b,], tunes=[1, 2], pols=[0, 1], nSegments=2)
        
        # Go
        while True:
            try:
                cFrame = drx.readFrame(fh)
            except errors.eofError:
                break
            except errors.syncError:
                continue

            frameBuffer.append(cFrame)
            cFrames = frameBuffer.get()

            if cFrames is None:
                continue
                
        fh.close()
        
        # Flush the buffer
        for cFrames in frameBuffer.flush():
            # Make sure the dump has one of the expected time tags
            self.assertTrue(cFrames[0].data.timeTag in (257355782095346056,))
            
            # Make sure it has the right number of frames
            self.assertEqual(len(cFrames), 4)
            
    def test_drx_buffer_reorder_flush(self):
        """Test the DRX ring buffer's flush() function with reordering."""
        
        fh = open(drxFile, 'rb')
        junkFrame = drx.readFrame(fh)
        b,t,p = junkFrame.parseID()
        fh.seek(-drx.FrameSize, 1)
        
        # Create the FrameBuffer instance
        frameBuffer = buffer.DRXFrameBuffer(beams=[b,], tunes=[1, 2], pols=[0, 1], nSegments=2, ReorderFrames=True)
        
        # Go
        while True:
            try:
                cFrame = drx.readFrame(fh)
            except errors.eofError:
                break
            except errors.syncError:
                continue

            frameBuffer.append(cFrame)
            cFrames = frameBuffer.get()

            if cFrames is None:
                continue
                
            # Make sure it has the right number of frames
            self.assertEqual(len(cFrames), 4)
            
            # Check the order
            for i in xrange(1, len(cFrames)):
                pB, pT, pP = cFrames[i-1].parseID()
                cB, cT, cP = cFrames[i].parseID()
                
                pID = 4*pB + 2*(pT-1) + pP
                cID = 4*cB + 2*(cT-1) + cP
                self.assertTrue(cID > pID)
                
        fh.close()
        
        # Flush the buffer
        for cFrames in frameBuffer.flush():
            # Make sure the dump has one of the expected time tags
            self.assertTrue(cFrames[0].data.timeTag in (257355782095346056,))
            
            # Make sure it has the right number of frames
            self.assertEqual(len(cFrames), 4)
            
            # Check the order
            for i in xrange(1, len(cFrames)):
                pB, pT, pP = cFrames[i-1].parseID()
                cB, cT, cP = cFrames[i].parseID()
                
                pID = 4*pB + 2*(pT-1) + pP
                cID = 4*cB + 2*(cT-1) + cP
                self.assertTrue(cID > pID)
                
    #
    # TBF
    #
    
    def test_tbf_default(self):
        """Test the TBF ring buffer with the default values."""
        
        fh = open(tbfFile, 'rb')
        nFpO = tbf.getFramesPerObs(fh)
        
        # Create the FrameBuffer instance
        frameBuffer = buffer.TBFFrameBuffer(chans=[2348, 2360, 2372, 2384, 2396])
        
        # Go
        while True:
            try:
                cFrame = tbf.readFrame(fh)
            except errors.eofError:
                break
            except errors.syncError:
                continue
                
            frameBuffer.append(cFrame)
            cFrames = frameBuffer.get()
            
            if cFrames is None:
                continue
            
        fh.close()
        
        # Make sure we have the right number of frames in the buffer
        nFrames = 0
        for key in frameBuffer.buffer.keys():
            nFrames = nFrames + len(frameBuffer.buffer[key])
        self.assertEqual(nFrames, 5)
        
        # Make sure nothing has happened that shouldn't have
        self.assertEqual(frameBuffer.full,    0)
        self.assertEqual(frameBuffer.partial, 0)
        self.assertEqual(frameBuffer.missing, 0)
        self.assertEqual(frameBuffer.dropped, 0)
        
        # Make sure we have the right keys
        for key in frameBuffer.buffer.keys():
            self.assertTrue(key in (283685766952000000,))
            
        # Make sure the buffer keys have the right sizes
        self.assertEqual(len(frameBuffer.buffer[283685766952000000]), 5)
        
    def test_tbf_small(self):
        """Test a small version of the TBF ring buffer."""
        
        fh = open(tbfFile, 'rb')
        nFpO = tbf.getFramesPerObs(fh)
        
        # Create the FrameBuffer instance
        frameBuffer = buffer.TBFFrameBuffer(chans=[2348, 2360, 2372, 2384, 2396], nSegments=1)
        
        # Go
        while True:
            try:
                cFrame = tbf.readFrame(fh)
            except errors.eofError:
                break
            except errors.syncError:
                continue
                
            frameBuffer.append(cFrame)
            cFrames = frameBuffer.get()
            
            if cFrames is None:
                continue
                
            # Make sure the dump has one of the expected time tags
            self.assertTrue(cFrames[0].data.timeTag in (283685766952000000,))
            
            # Make sure it has the right number of frames
            self.assertEqual(len(cFrames), nFpO)
            
        fh.close()
        
    def test_tbf_buffer_reorder(self):
        """Test the reorder function of the TBF ring buffer."""
        
        fh = open(tbfFile, 'rb')
        nFpO = tbf.getFramesPerObs(fh)
        
        # Create the FrameBuffer instance
        frameBuffer = buffer.TBFFrameBuffer(chans=[2348, 2360, 2372, 2384, 2396], ReorderFrames=True)
        
        # Go
        while True:
            try:
                cFrame = tbf.readFrame(fh)
            except errors.eofError:
                break
            except errors.syncError:
                continue
                
            frameBuffer.append(cFrame)
            cFrames = frameBuffer.get()
            
            if cFrames is None:
                continue
                
            # Make sure the dump has one of the expected time tags
            self.assertTrue(cFrames[0].data.timeTag in (283685766952000000,))
            
            # Make sure it has the right number of frames
            self.assertEqual(len(cFrames), nFpO)
            
            # Check the order
            for i in xrange(1, len(cFrames)):
                pC = cFrames[i-1].header.firstChan
                cC = cFrames[i].header.firstChan
                
                self.assertTrue(cC > pC)
                
        fh.close()
        
    def test_tbf_buffer_flush(self):
        """Test the TBF ring buffer's flush() function."""
        
        fh = open(tbfFile, 'rb')
        nFpO = tbf.getFramesPerObs(fh)
        
        # Create the FrameBuffer instance
        frameBuffer = buffer.TBFFrameBuffer(chans=[2348, 2360, 2372, 2384, 2396])
        
        # Go
        while True:
            try:
                cFrame = tbf.readFrame(fh)
            except errors.eofError:
                break
            except errors.syncError:
                continue
                
            frameBuffer.append(cFrame)
            cFrames = frameBuffer.get()
            
            if cFrames is None:
                continue
                
        fh.close()
        
        # Flush the buffer
        for cFrames in frameBuffer.flush():
            # Make sure the dump has one of the expected time tags
            self.assertTrue(cFrames[0].data.timeTag in (283685766952000000,))
            
            # Make sure it has the right number of frames
            self.assertEqual(len(cFrames), nFpO)
            
    def test_tbf_buffer_reorder_flush(self):
        """Test the TBF ring buffer's flush() function with reordering."""
        
        fh = open(tbfFile, 'rb')
        nFpO = tbf.getFramesPerObs(fh)
        
        # Create the FrameBuffer instance
        frameBuffer = buffer.TBFFrameBuffer(chans=[2348, 2360, 2372, 2384, 2396], ReorderFrames=True)
        
        # Go
        while True:
            try:
                cFrame = tbf.readFrame(fh)
            except errors.eofError:
                break
            except errors.syncError:
                continue
                
            frameBuffer.append(cFrame)
            cFrames = frameBuffer.get()
            
            if cFrames is None:
                continue
                
        fh.close()
        
        # Flush the buffer
        for cFrames in frameBuffer.flush():
            # Make sure the dump has one of the expected time tags
            self.assertTrue(cFrames[0].data.timeTag in (283685766952000000,))
            
            # Make sure it has the right number of frames
            self.assertEqual(len(cFrames), nFpO)
            
            # Check the order
            for i in xrange(1, len(cFrames)):
                pC = cFrames[i-1].header.firstChan
                cC = cFrames[i].header.firstChan
                
                self.assertTrue(cC > pC)


class buffer_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.reader.buffer 
    unit tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(buffer_tests)) 


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Unit test for lsl.common.paths module."""

import os
import unittest

from lsl.common import paths


__revision__ = "$Rev: 2394 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


class paths_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.common.paths tests
    module."""

    def test_module_path(self):
        """Test the paths.module variable."""
        
        modPath = paths.moduleBuild

        astroFile = os.path.join(modPath, 'astro.py')
        self.assertTrue(os.path.exists(astroFile))

        tbwFile = os.path.join(modPath, 'reader', 'tbw.py')
        self.assertTrue(os.path.exists(tbwFile))

    def test_data_path(self):
        """Test the paths.data variable."""
        
        dataPath = paths.dataBuild

        ssmif = os.path.join(dataPath, 'lwa1-ssmif.txt')
        self.assertTrue(os.path.exists(ssmif))

        timeFile = os.path.join(dataPath, 'astro', 'tai-utc.dat')
        self.assertTrue(os.path.exists(timeFile))

class paths_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.common.paths
    module unit tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(paths_tests)) 


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Unit test for lsl.misc.geodesy module."""

import os
import unittest

from lsl.misc import geodesy


__revision__ = "$Rev: 2394 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


class geodesy_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.misc.geodesy module."""
    
    def test_read_MAIA(self):
        """Test that a line from one of the MAIA files can be read."""
        
        ACCURACY = 6
        
        ## http://toshi.nofs.navy.mil/ser7/finals2000A.all
        line = '92 1 1 48622.00 I  0.182985 0.000672  0.168782 0.000345  I-0.1251614 0.0000207  1.8335 0.0201  I    -0.086    0.202     0.130    0.165   .182400   .167900  -.1253000     0.129    -0.653\n'
        newEOP = geodesy.EOP()
        newEOP.fromMAIA(line)
        
        self.assertAlmostEqual(newEOP.x, 0.182985, ACCURACY)
        self.assertAlmostEqual(newEOP.y, 0.168782, ACCURACY)
        self.assertAlmostEqual(newEOP.utDiff, -0.1251614, ACCURACY)
        self.assertEqual(newEOP.type, 'final')
        
    def test_read_mjd(self):
        """Test reading a specific MJD via getEOP."""
        
        ACCURACY = 6
        
        # Read in the data for January 2, 1973
        eop = geodesy.getEOP(41684.0)
        
        # Check
        ## http://toshi.nofs.navy.mil/ser7/finals2000A.all
        ## 73 1 2 41684.00 I  0.120733 0.009786  0.136966 0.015902  I 0.8084178 0.0002710  0.0000 0.1916  P    -0.766    0.199    -0.720    0.300   .143000   .137000   .8075000   -18.637    -3.667 
        self.assertAlmostEqual(eop.x, 0.120733, ACCURACY)
        self.assertAlmostEqual(eop.y, 0.136966, ACCURACY)
        self.assertAlmostEqual(eop.utDiff,  0.8084178, ACCURACY)
        
        # Read in the data for January 2, 1993
        eops = geodesy.getEOP(48989.0)
        eop = eops
        
        # Check
        ## http://toshi.nofs.navy.mil/ser7/finals2000A.all
        ## 93 1 2 48989.00 I  0.208333 0.000166  0.356577 0.000241  I 0.0594754 0.0000108  2.6575 0.0094  I     0.187    0.233     0.020    0.180   .208700   .355900   .0593100    -0.020    -0.087 
        self.assertAlmostEqual(eop.x, 0.208333, ACCURACY)
        self.assertAlmostEqual(eop.y, 0.356577, ACCURACY)
        self.assertAlmostEqual(eop.utDiff,  0.0594754, ACCURACY)
        
        # Read in three values and check the output
        mjdList = [41684.0, 41685.0, 41686.0]
        eops = geodesy.getEOP(mjdList)
        self.assertEqual(len(eops), 3)
        for i in range(3):
            self.assertTrue(mjdList[i] in eops)


class geodesy_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.misc.geodesy
    module unit tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(geodesy_tests)) 


if __name__ == '__main__':
    unittest.main()




# -*- coding: utf-8 -*

"""lsl.reader - Modular readers for the various LWA data formats:
 * tbw
 * tbn
 * drx
 * drspec
 * vdif
 * tbf
 * cor

A ring buffer for re-ordering TBN data is included in the 'buffer'
module.

Also include as part of this module are the LWA Development
Primities (LDP), a collection of utilites that make reading in
data files fun.
"""




# -*- coding: utf-8 -*-

# Python3 compatiability
import sys
if sys.version_info > (3,):
    xrange = range
    
"""
Module to support imaging correlated data.  This module provides utilities to 
read FITS IDI files into data dictionaries (as described in 
:mod:`lsl.sim.vis`) and build AIPY ImgW instances from the data dictionaries.
Also included is a utility to sort data dictionaries by baselines.

.. versionadded:: 0.5.0

.. versionchanged:: 1.0.0
    Added support for UVFITS files and CASA measurement sets

.. versionchanged:: 1.0.1
    Added the plotGriddedImage() function
    
.. versionchanged:: 1.1.0
    Added the getImageRADec() and getImageAzEl() functions to complement
    plotGriddedImage() and make it easier to work with phase centers 
    that are not at zenith.  Added in the ImgWPlus class to add support
    for imaging weighting and tapering.
    
.. versionchanged:: 1.1.2
    Added support for CASA MeasurementSets that are stored as a tarball
    
.. versionchanged:: 1.1.4
    Fixed a conjugation problem in the visibilities read from a FITS-IDI file
"""

import os
import re
import sys
import aipy
import ephem
import numpy
import atexit
import pyfits
import shutil
import tarfile
import tempfile
try:
    import cStringIO as StringIO
except ImportError:
    import StringIO
from calendar import timegm
from datetime import datetime
from operator import itemgetter

from lsl import astro
from lsl.statistics import robust
from lsl.common import stations
from lsl.sim import vis as simVis
from lsl.writer.fitsidi import NumericStokes
from lsl.writer.measurementset import NumericStokes as NumericStokesMS
from lsl.common.constants import c as vLight

try:
    import pyfftw
    
    # Enable the PyFFTW cache
    if not pyfftw.interfaces.cache.is_enabled():
        pyfftw.interfaces.cache.enable()
        pyfftw.interfaces.cache.set_keepalive_time(60)
        
    fft2Function = lambda x: pyfftw.interfaces.numpy_fft.fft2(x)
    ifft2Function = lambda x: pyfftw.interfaces.numpy_fft.ifft2(x)

except ImportError:
    from scipy import fftpack
    fft2Function = fftpack.fft2
    ifft2Function = fftpack.ifft2

__version__ = '0.8'
__revision__ = '$Rev: 2598 $'
__all__ = ['baselineOrder', 'sortDataDict', 'pruneBaselineRange', 'rephaseData', 'CorrelatedData', 
           'CorrelatedDataIDI', 'CorrelatedDataUV', 'CorrelatedDataMS', 'ImgWPlus', 'buildGriddedImage', 
           'plotGriddedImage', 'getImageRADec', 'getImageAzEl', '__version__', '__revision__', '__all__']


# Regular expression for trying to get the stand number out of an antenna
# name
_annameRE = re.compile('^.*?(?P<id>\d{1,3})$')


def baselineOrder(bls):
    """
    Like numpy.argsort(), but for a list of two-element tuples of baseline 
    pairs.  The resulting lists can then be used to sort a data dictionary
    a la sortDataDict().
    """
    
    def __cmpBaseline(bl):
        return 1024*bl[0] + bl[1]
    
    return [i for (v, i) in sorted((v, i) for (i, v) in enumerate([__cmpBaseline(bl) for bl in bls]))]


def sortDataDict(dataDict, order=None):
    """
    Sort a data dictionary by the specified order.  If no order is supplied, 
    the data dictionary is sorted by baseline using baselineOrder().
    """
    
    if order is None:
        for pol in ['xx', 'yy', 'rr', 'll', 'I']:
            try:
                if len(dataDict['bls'][pol]) == 0:
                    continue
                order = baselineOrder(dataDict['bls'][pol])
                break
            except KeyError:
                pass
    
    for key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
        for pol in dataDict[key].keys():
            try:
                newList = [dataDict[key][pol][i] for i in order]
                dataDict[key][pol] = newList
            except (KeyError, IndexError):
                pass
            
    return dataDict


def pruneBaselineRange(dataDict, uvMin=0, uvMax=numpy.inf):
    """
    Prune baselines from a data dictionary that are less than uvMin or
    greater than or equal to uvMax.

    .. note::
        uvMin and uvMax should be specified in lambda
    """

    # Force min to be less than max
    if uvMin > uvMax:
        temp = uvMin
        uvMin = uvMax
        uvMax = temp
        
    # Create the new output data dictionary
    newDict = {}
    for key in dataDict.keys():
        if key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            newDict[key] = {}
            for pol in dataDict[key].keys():
                newDict[key][pol] = []
        else:
            newDict[key] = dataDict[key]

    # Find out the baseline lengths and create a list of good ones
    good = {}
    freq = dataDict['freq']
    for pol in dataDict['uvw'].keys():
        sizes = []
        for bl in dataDict['uvw'][pol]:
            uvw = bl[:,freq.size/2]
            sizes.append( numpy.sqrt((uvw**2).sum()) )
        sizes = numpy.array(sizes)
        good[pol] = list(numpy.where( (sizes >= uvMin) & (sizes < uvMax) )[0])
        
    # Prune
    for key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
        for pol in dataDict[key].keys():
            lgp = len(good[pol])
            if lgp == 0:
                newDict[key][pol] = []
            elif lgp == 1:
                newDict[key][pol] = [dataDict[key][pol][good[pol][0]],]
            else:
                newDict[key][pol] = list(itemgetter(*good[pol])(dataDict[key][pol]))
                
    # Return
    return newDict


def rephaseData(aa, dataDict, currentPhaseCenter='z', newPhaseCenter='z'):
    """
    Given an AntennaArray instance and a data dictionary, re-phase the data 
    to change the pointing center.
    """
    
    # Load in basic information about the data
    freq = dataDict['freq']*1.0
    isMasked = dataDict['isMasked']
    pols = dataDict['bls'].keys()
    
    # Create the data dictionary that will hold the re-phased data
    dataDict2 = {}
    dataDict2['freq'] = freq 
    dataDict2['isMasked'] = isMasked
    for key in ('uvw', 'vis', 'wgt', 'msk', 'bls', 'jd'):
        dataDict2[key] = {}
        for p in pols:
            dataDict2[key][p] = []
            
    # Go!
    for p in pols:
        lastJD = None
        
        ## Loop over baselines
        for k in xrange(len(dataDict['bls'][p])):
            ### Load in the data
            i,j = dataDict['bls'][p][k]
            d   = dataDict['vis'][p][k]
            msk = dataDict['msk'][p][k]
            jd  = dataDict[ 'jd'][p][k]
            
            ### Update the source positions if needed
            if jd != lastJD:
                # Update the time for the AntennaArray
                aa.set_jultime(jd)
                
                # Recompute
                if currentPhaseCenter is not 'z':
                    currentPhaseCenter.compute(aa)
                if newPhaseCenter is not 'z':
                    newPhaseCenter.compute(aa)
                    
                lastJD = jd
                
            ### Compute the uvw coordinates and the new phasing
            try:
                crd = aa.gen_uvw(j, i, src=newPhaseCenter)[:,0,:]
                d = aa.unphs2src(d, currentPhaseCenter, j, i)
                d = aa.phs2src(d, newPhaseCenter, j, i)
            except aipy.phs.PointingError:
                raise RuntimeError("Rephasing center is below the horizon")
                
            ### Save
            if isMasked:
                crd = crd.compress(numpy.logical_not(msk), axis=2)
            vis = d.compress(numpy.logical_not(msk))
            wgt = numpy.ones_like(vis) * len(vis)

            dataDict2['bls'][p].append( (i,j) )
            dataDict2['uvw'][p].append( crd )
            dataDict2['vis'][p].append( vis )
            dataDict2['wgt'][p].append( wgt )
            dataDict2['msk'][p].append( msk )
            dataDict2[ 'jd'][p].append( jd )
            
    # Done
    return dataDict2


def CorrelatedData(filename):
    """
    Read in and work with FITS IDI and UVFITS files.  Returns either a 
    CorrelateDataIDI or CorrelatedDataUV instance.
    """
    
    valid = False
    
    # Basic filesystem validation
    if not os.path.exists(filename):
        raise IOError("File does not exists")
    if not os.access(filename, os.R_OK):
        raise IOError("File cannot be read")
        
    if os.path.isdir(filename):
        # Only a MS can be a directory
        try:
            return CorrelatedDataMS(filename)
        except Exception as e:
            pass
            
    else:
        # Standard files
        ## FITS IDI
        try:
            return CorrelatedDataIDI(filename)
        except Exception as e:
            pass
        
        ## UVFITS
        try:
            return CorrelatedDataUV(filename)
        except Exception as e:
            pass
        
        ## Measurment Set as a compressed entity
        try:
            return CorrelatedDataMS(filename)
        except Exception as e:
            pass
        
    if not valid:
        raise RuntimeError("File '%s' does not appear to be either a FITS IDI file, UV FITS file, or MeasurmentSet" % filename)


class CorrelatedDataIDI(object):
    """
    Class to make accessing information about a FITS IDI easy.  This wraps 
    all of the "messy" machinery needed to extract both the metadata and data 
    from the file and return them as common LSL objects.
    
    This class has three main attributes to interact with:
      * getAntennaArray - Return a :class:`lsl.sim.vim.AntennaArray` instance
                          that represents the array where the data was obtained.  
                          This is useful for simulation proposes and computing 
                          source positions.
      * getObserver - Return a ephem.Observer instance representing the array
      * getDataSet - Return a data dictionary of all baselines for a given set
                     of observations
        
    The class also includes a variety of useful metadata attributes:
      * pols - Numpy array of polarization product codes
      * freq - Numpy array of frequency channels in Hz
      * station - LSL :class:`lsl.common.stations.LWAStation` instance for the
                  array
      * dateObs - Datetime object for the reference date of the FIT IDI file
      * antennas - List of :class:`lsl.common.stations.Antenna` instances
    
    .. note::
        The CorrelatedData.antennas attribute should be used over 
        CorrelatedData.station.getAntennas() since the mapping in the FITS IDI
        file may not be the same as the digitizer order.
    """
    
    def _createEmptyDataDict(self):
        """
        Create an empty data dictionary that is appropriate for the current file.
        """
        
        dataDict = {}
        dataDict['freq'] = self.freq 
        dataDict['isMasked'] = False
        for key in ('uvw', 'vis', 'wgt', 'msk', 'bls', 'jd'):
            dataDict[key] = {}
            
        for p in self.pols:
            name = NumericStokes[p]
            if len(name) == 2:
                name = name.lower()
            for key in ('uvw', 'vis', 'wgt', 'msk', 'bls', 'jd'):
                dataDict[key][name] = []
                
        return dataDict
    
    def __init__(self, filename):
        """
        Initialize a new CorrelatedDataIDI instance from a FITS IDI file and 
        fill in the metadata.
        """
        
        self.filename = filename
        
        # Open the file, check if it looks like FITS IDI, and pull out the UV_DATA table
        hdulist = pyfits.open(self.filename)
        tbls = [i.header['EXTNAME'] for i in hdulist[1:]]
        for tbl in ('ARRAY_GEOMETRY', 'FREQUENCY', 'ANTENNA', 'SOURCE', 'UV_DATA'):
            if tbl not in tbls:
                raise RuntimeError("Cannot find table '%s' in '%s'" % (tbl, self.filename))
        
        self.extended = False
        self.conjugate = True
        try:
            if hdulist[0].header['LWATYPE'] == 'IDI-EXTENDED-ZA':
                self.extended = True
        except KeyError:
            ## Catch for LEDA64-NM data
            pass
        try:
            if hdulist[0].header['LWAMAJV'] < 3:
                self.conjugate = False
        except KeyError:
            pass
        ag = hdulist['ARRAY_GEOMETRY']
        fq = hdulist['FREQUENCY']
        uvData = hdulist['UV_DATA']
        
        # Antennas
        try:
            mapper = hdulist['NOSTA_MAPPER']
            
            nosta = mapper.data.field('NOSTA')
            noact = mapper.data.field('NOACT')
            stabxyz = ag.data.field('STABXYZ')
            anname = mapper.data.field('ANNAME')
        except KeyError:
            nosta = ag.data.field('NOSTA')
            noact = ag.data.field('NOSTA')
            stabxyz = ag.data.field('STABXYZ')
            anname = ag.data.field('ANNAME')
        
        # Station/telescope information
        try:
            self.telescope = hdulist[0].header['TELESCOP']
            self.dateObs = datetime.strptime(hdulist[0].header['DATE-OBS'], "%Y-%m-%dT%H:%M:%S")
        except ValueError:
            ## Catch for DiFX FITS-IDI files
            self.dateObs = datetime.strptime(hdulist[0].header['DATE-OBS'], "%Y-%m-%d")
        except KeyError:
            ## Catch for LEDA64-NM data
            self.telescope = uvData.header['TELESCOP']
            self.dateObs = datetime.strptime(uvData.header['DATE-OBS'], "%Y-%m-%dT%H:%M:%S")
            
        ## Extract the site position
        geo = numpy.array([ag.header['ARRAYX'], ag.header['ARRAYY'], ag.header['ARRAYZ']])
        site = stations.ecef2geo(*geo)
        
        ## Try to back out the "real" stand names
        noact2 = []
        for nam in anname:
            try:
                mtch =  _annameRE.match(nam)
                id = int(mtch.group('id'))
                noact2.append(id)
            except (ValueError, AttributeError):
                break
        if len(noact2) == len(noact):
            noact = numpy.array(noact2)
            
        ## Create the ECI -> topocentric transform
        lat  = site[0]
        ecii = numpy.array([[ 0.0,            1.0, 0.0           ],
                        [-numpy.sin(lat), 0.0, numpy.cos(lat)],
                        [ numpy.cos(lat), 0.0, numpy.sin(lat)]])
                        
        ## Build up the list of antennas
        antennas = []
        for line,act in zip(ag.data, noact):
            enz = numpy.dot(ecii, line['STABXYZ'])
            
            stand = stations.Stand(act, *enz)
            antennas.append(stations.Antenna(2*(stand.id-1)+1, stand=stand, pol=0))
            
        ## Build up the station
        self.station = stations.LWAStation(ag.header['ARRNAM'], site[0]*180/numpy.pi, site[1]*180/numpy.pi, site[2], antennas=antennas)
        
        self.standMap = {}
        self.stands = []
        for sta, act in zip(nosta, noact):
            self.standMap[sta] = act
            self.stands.append(act)
            
        self.antennaMap = {}
        self.antennas = []
        for stand in self.stands:
            for ant in self.station.getAntennas():
                if ant.stand.id == stand and ant.pol == 0:
                    self.antennas.append(ant)
                    self.antennaMap[ant.stand.id] = ant
                    break
                    
        # Polarization and frequency
        self.pols  = numpy.arange(1, uvData.header['MAXIS2']+1) - uvData.header['CRPIX2']
        self.pols *= uvData.header['CDELT2'] 
        self.pols += uvData.header['CRVAL2']
        self.freq  = numpy.array([], dtype=numpy.float64)
        for i in xrange(uvData.header['NO_BAND']):
            width = fq.data['CH_WIDTH'][0][i] if uvData.header['NO_BAND'] > 1 else fq.data['CH_WIDTH'][0]
            offset = fq.data['BANDFREQ'][0][i] if uvData.header['NO_BAND'] > 1 else fq.data['BANDFREQ'][0]
            
            freqIF = numpy.arange(1, uvData.header['NO_CHAN']+1, dtype=numpy.float64) - uvData.header['REF_PIXL']
            freqIF *= width
            freqIF += uvData.header['REF_FREQ'] + offset
            self.freq = numpy.concatenate([self.freq, freqIF])
            
        # Total baseline count
        self.totalBaselineCount = len(uvData.data['BASELINE'])
        
        # Integration count
        jd = uvData.data['DATE'] + uvData.data['TIME']
        self.integrationCount = len(numpy.unique(jd))
        
        # Close
        hdulist.close()
    
    def getAntennaArray(self):
        """
        Return an AIPY AntennaArray instance for the array that made the 
        observations contained here.
        """
        
        # Get the date of observations
        refJD = astro.unix_to_utcjd(timegm(self.dateObs.timetuple()))
        
        # Return
        return simVis.buildSimArray(self.station, self.antennas, self.freq/1e9, jd=refJD)
        
    def getObserver(self):
        """
        Return a ephem.Observer instances for the array described in the file.
        """
        
        # Get the date of observations
        refJD = astro.unix_to_utcjd(timegm(self.dateObs.timetuple()))
        
        obs = self.station.getObserver()
        obs.date = refJD - astro.DJD_OFFSET
        return obs
        
    def getDataSet(self, set, includeAuto=False, sort=True, uvMin=0, uvMax=numpy.inf):
        """
        Return a baseline sorted data dictionary for the specified data set.  
        By default this excludes the autocorrelations.  To include 
        autocorrelations set the value of 'includeAuto' to True.  Setting the
        'sort' keyword to False will disable the baseline sorting.  Optionally,
        baselines with lengths between uvMin and uvMax can only be returned.

        .. note::
            uvMin and uvMax should be specified in lambda
            
        .. versionchanged:: 1.1.0
            'set' can now be either an integer or a list to pull back multiple 
            integrations.
        """
        
        # Open the file
        hdulist = pyfits.open(self.filename)
        uvData = hdulist['UV_DATA']
        
        # We need this a lot...
        nPol = len(self.pols)
        
        # Define the dictionary to return
        dataDict = self._createEmptyDataDict()

        # Set the source ID to look for (this is LWA specific)
        if type(set) == list:
            sourceID = set
        else:
            sourceID = range(set,set+1)
            
        # Figure out what rows we need based on how the baseline iterate
        blList = uvData.data['BASELINE']
        ## Baseline based boundaries
        setBoundaries = numpy.where( blList == blList[0] )[0]
        try:
            setStart = setBoundaries[ sourceID[ 0]-1 ]
        except IndexError:
            raise RuntimeError("Cannot find baseline set %i in FITS IDI file", set)
        try:
            setStop  = setBoundaries[ sourceID[-1]   ]
        except IndexError:
            setStop = len(blList)
        ## Row Selection
        selection = numpy.s_[setStart:setStop]
        
        # Figure out if we have seperate WEIGHT data or not
        seperateWeights = False
        for col in uvData.data.columns:
            if col.name == 'WEIGHT':
                seperateWeights = True
                break
                
        # Pull out the raw data from the table
        bl = blList[selection]
        jd = uvData.data['DATE'][selection] + uvData.data['TIME'][selection]
        try:
            u, v, w = uvData.data['UU'][selection], uvData.data['VV'][selection], uvData.data['WW'][selection]
        except KeyError:
            u, v, w = uvData.data['UU---SIN'][selection], uvData.data['VV---SIN'][selection], uvData.data['WW---SIN'][selection]
        vis = numpy.ascontiguousarray(uvData.data['FLUX'][selection])
        if seperateWeights:
            wgt = numpy.ascontiguousarray(uvData.data['WEIGHT'][selection])
        else:
            wgt = None
            
        # Re-work the data into something more useful
        ## Axis sizes
        nFreq = len(self.freq)
        nStk = len(self.pols)
        nCmp = 2 if seperateWeights else 3
        if vis.size/nFreq/nStk/nCmp != len(bl):
            ### Catch for FITS-IDI files generate by interfits
            nCmp = 2 if nCmp == 3 else 3
        ## Frequency for converting the u, v, and w coordinates
        freq = self.freq*1.0
        freq.shape += (1,)
        ## Convert u, v, and w from seconds to wavelengths and then into one massive array
        u = (u*freq).T
        v = (v*freq).T
        w = (w*freq).T
        uvw = numpy.array([u,v,w], dtype=numpy.float32)
        ## Reshape the visibilities and weights
        vis.shape = (vis.size/nFreq/nStk/nCmp, nFreq, nStk, nCmp)
        if seperateWeights:
            if wgt.shape != nFreq*nStk:
                ## Catch for some old stuff
                wgt = numpy.concatenate([wgt for pol in self.pols])
            wgt.shape = (wgt.size/nFreq/nStk, nFreq, nStk)
        else:
            try:
                wgt = vis[:,:,:,2]
                vis = vis[:,:,:,:2]
            except IndexError:
                ### Catch for FITS-IDI files generate by interfits
                wgt = numpy.ones([vis.shape[i] for i in xrange(3)], dtype=numpy.float32)
        ## Back to complex
        vis = vis[:,:,:,0] + 1j*vis[:,:,:,1]
        if self.conjugate:
            ## NOTE: This is this conjugate since there seems to be a convention mis-match
            ##       between LSL and AIPS/the FITS-IDI convention.
            vis = vis.conj()
        ## Scale
        try:
            scl = uvData.header['VIS_SCAL']
            vis /= scl
        except KeyError:
            pass
            
        ## Setup a dummy mask
        msk = numpy.zeros(nFreq, dtype=numpy.int16)
        
        # Re-pack into the data dictionary
        for b in xrange(bl.size):
            if not self.extended:
                i = self.standMap[(bl[b] >> 8) & 255]
                j = self.standMap[bl[b] & 255]
            else:
                i = self.standMap[(bl[b] >> 16) & 65535]
                j = self.standMap[bl[b] & 65535]
            if i == j and not includeAuto:
                ## Skip auto-correlations
                continue
            ri = numpy.where(self.stands == i)[0][0]
            rj = numpy.where(self.stands == j)[0][0]
            
            for p,l in enumerate(self.pols):
                name = NumericStokes[l]
                if len(name) == 2:
                    name = name.lower()
                    
                dataDict['bls'][name].append( (ri,rj) )
                dataDict['uvw'][name].append( uvw[:,b,:] )
                dataDict['vis'][name].append( vis[b,:,p] )
                dataDict['wgt'][name].append( wgt[b,:,p] )
                dataDict['msk'][name].append( msk )
                dataDict['jd' ][name].append( jd[b] )
                
        # Close
        hdulist.close()
        
        # Sort
        if sort:
            sortDataDict(dataDict)
            
        # Prune
        if uvMin != 0 or uvMax != numpy.inf:
            dataDict = pruneBaselineRange(dataDict, uvMin=uvMin, uvMax=uvMax)
            
        # Return
        return dataDict


class CorrelatedDataUV(object):
    """
    Class to make accessing information about a UVFITS file easy.  This wraps 
    all of the "messy" machinery needed to extract both the metadata and data 
    from the file and return them as common LSL objects.
    
    This class has three main attributes to interact with:
      * getAntennaArray - Return a :class:`lsl.sim.vim.AntennaArray` instance
                          that represents the array where the data was obtained.  
                          This is useful for simulation proposes and computing 
                          source positions.
      * getObserver - Return a ephem.Observer instance representing the array
      * getDataSet - Return a data dictionary of all baselines for a given set
                     of observations
        
    The class also includes a variety of useful metadata attributes:
      * pols - Numpy array of polarization product codes
      * freq - Numpy array of frequency channels in Hz
      * station - LSL :class:`lsl.common.stations.LWAStation` instance for the
                  array
      * dateObs - Datetime object for the reference date of the FIT IDI file
      * antennas - List of :class:`lsl.common.stations.Antenna` instances
    
    .. note::
        The CorrelatedDataUV.antennas attribute should be used over 
        CorrelatedDataUV.station.getAntennas() since the mapping in the UVFITS
        file may not be the same as the digitizer order.
    """
    
    def _createEmptyDataDict(self):
        """
        Create an empty data dictionary that is appropriate for the current file.
        """
        
        dataDict = {}
        dataDict['freq'] = self.freq 
        dataDict['isMasked'] = False
        for key in ('uvw', 'vis', 'wgt', 'msk', 'bls', 'jd'):
            dataDict[key] = {}
            
        for p in self.pols:
            name = NumericStokes[p]
            if len(name) == 2:
                name = name.lower()
            for key in ('uvw', 'vis', 'wgt', 'msk', 'bls', 'jd'):
                dataDict[key][name] = []
                
        return dataDict
    
    def __init__(self, filename):
        """
        Initialize a new CorrelatedDataUV instance from a UVFITS file and 
        fill in the metadata.
        """
        
        self.filename = filename
        
        # Open the various tables that we need
        hdulist = pyfits.open(filename)
        
        uvData = hdulist[0]
        ag = hdulist['AIPS AN']
        
        # Antennas
        nosta = ag.data.field('NOSTA')
        noact = ag.data.field('NOSTA')
        anname = ag.data.field('ANNAME')
        
        # Station/telescope information
        self.telescope = hdulist[0].header['TELESCOP']
        dt = hdulist[0].header['DATE-OBS']
        dt = dt.rsplit('.', 1)[0]
        try:
            self.dateObs = datetime.strptime(dt, "%Y-%m-%dT%H:%M:%S")
        except ValueError:
            ## Catch for AIPS UVFITS files which only have a date set
            self.dateObs = datetime.strptime(dt, "%Y-%m-%d")
            
        ## Extract the site position
        geo = numpy.array([ag.header['ARRAYX'], ag.header['ARRAYY'], ag.header['ARRAYZ']])
        site = stations.ecef2geo(*geo)
        
        ## Try to back out the "real" stand names
        noact2 = []
        for nam in anname:
            try:
                mtch =  _annameRE.match(nam)
                id = int(mtch.group('id'))
                noact2.append(id)
            except (ValueError, AttributeError):
                break
        if len(noact2) == len(noact):
            noact = numpy.array(noact2)
            
        ## Create the ECI -> topocentric transform
        lat  = site[0]
        ecii = numpy.array([[ 0.0,            1.0, 0.0           ],
                        [-numpy.sin(lat), 0.0, numpy.cos(lat)],
                        [ numpy.cos(lat), 0.0, numpy.sin(lat)]])
                        
        ## Build up the list of antennas
        antennas = []
        for line,act in zip(ag.data, noact):
            enz = numpy.dot(ecii, line['STABXYZ'])
            
            stand = stations.Stand(act, *enz)
            antennas.append(stations.Antenna(2*(stand.id-1)+1, stand=stand, pol=0))
            
        ## Build up the station
        self.station = stations.LWAStation(ag.header['ARRNAM'], site[0]*180/numpy.pi, site[1]*180/numpy.pi, site[2], antennas=antennas)
        
        self.standMap = {}
        self.stands = []
        for nosta, noact in zip(nosta, noact):
            self.standMap[nosta] = noact
            self.stands.append(noact)
            
        self.antennaMap = {}
        self.antennas = []
        for stand in self.stands:
            for ant in self.station.getAntennas():
                if ant.stand.id == stand and ant.pol == 0:
                    self.antennas.append(ant)
                    self.antennaMap[ant.stand.id] = ant
                    break
                    
        # Polarization and frequency
        self.pols  = numpy.arange(1, uvData.header['NAXIS3']+1) - uvData.header['CRPIX3']
        self.pols *= uvData.header['CDELT3'] 
        self.pols += uvData.header['CRVAL3']
        nChan = uvData.header['NAXIS4']
        if uvData.header['CTYPE5'] == 'IF':
            ## Merge the frequency and IF columns
            nChan *= uvData.header['NAXIS5']
        self.freq  = numpy.arange(1, nChan+1, dtype=numpy.float64) - uvData.header['CRPIX4']
        self.freq *= uvData.header['CDELT4']
        self.freq += uvData.header['CRVAL4']
        
        # Total baseline count
        self.totalBaselineCount = len(hdulist[0].data['BASELINE'])
        
        # Integration count
        jd = hdulist[0].data['DATE']
        self.integrationCount = len(numpy.unique(jd))
        
        # Close
        hdulist.close()
    
    def getAntennaArray(self):
        """
        Return an AIPY AntennaArray instance for the array that made the 
        observations contained here.
        """
        
        # Get the date of observations
        refJD = astro.unix_to_utcjd(timegm(self.dateObs.timetuple()))
        
        # Return
        return simVis.buildSimArray(self.station, self.antennas, self.freq/1e9, jd=refJD)
        
    def getObserver(self):
        """
        Return a ephem.Observer instances for the array described in the file.
        """
        
        # Get the date of observations
        refJD = astro.unix_to_utcjd(timegm(self.dateObs.timetuple()))
        
        obs = self.station.getObserver()
        obs.date = refJD - astro.DJD_OFFSET
        return obs
        
    def getDataSet(self, set, includeAuto=False, sort=True, uvMin=0, uvMax=numpy.inf):
        """
        Return a baseline sorted data dictionary for the specified data set.  
        By default this excludes the autocorrelations.  To include 
        autocorrelations set the value of 'includeAuto' to True.  Setting the
        'sort' keyword to False will disable the baseline sorting.  Optionally,
        baselines with lengths between uvMin and uvMax can only be returned.

        .. note::
            uvMin and uvMax should be specified in lambda
            
        .. versionchanged:: 1.1.0
            'set' can now be either an integer or a list to pull back multiple 
            integrations.
        """
        
        # Open the file
        hdulist = pyfits.open(self.filename)
        uvData = hdulist[0]
        
        # We need this a lot...
        nPol = len(self.pols)
        
        # Define the dictionary to return
        dataDict = self._createEmptyDataDict()

        # Set the source ID to look for (this is LWA specific)
        if type(set) == list:
            sourceID = set
        else:
            sourceID = range(set,set+1)
            
        # Figure out what rows we need based on how the baseline iterate
        blList = uvData.data['BASELINE']
        ## Baseline based boundaries
        setBoundaries = numpy.where( blList == blList[0] )[0]
        try:
            setStart = setBoundaries[ sourceID[ 0]-1 ]
        except IndexError:
            raise RuntimeError("Cannot find baseline set %i in FITS IDI file", set)
        try:
            setStop  = setBoundaries[ sourceID[-1]   ]
        except IndexError:
            setStop = len(blList)
        ## Row Selection
        selection = numpy.s_[setStart:setStop]

        # Pull out the raw data from the table
        bl = blList[selection]
        jd = uvData.data['DATE'][selection]
        try:
            u, v, w = uvData.data['UU'][selection], uvData.data['VV'][selection], uvData.data['WW'][selection]
        except KeyError:
            u, v, w = uvData.data['UU---SIN'][selection], uvData.data['VV---SIN'][selection], uvData.data['WW---SIN'][selection]
        vis = uvData.data['DATA'][selection]
        wgt = None
        
        # Re-work the data into something more useful
        ## Axis sizes
        nFreq = len(self.freq)
        nStk = len(self.pols)
        nCmp = vis.shape[-1]
        ## Frequency for converting the u, v, and w coordinates
        freq = self.freq*1.0
        freq.shape += (1,)
        ## Convert u, v, and w from seconds to wavelengths and then into one massive array
        u = (u*freq).T
        v = (v*freq).T
        w = (w*freq).T
        uvw = numpy.array([u,v,w], dtype=numpy.float32)
        ## Reshape the visibilities and weights
        if len(vis.shape) == 7:
            ### Merge the frequency and IF columns
            vis = vis[:,0,0,:,:,:,:]
            vis.shape = (vis.shape[0], vis.shape[1]*vis.shape[2], vis.shape[3], vis.shape[4])
        else:
            vis = vis[:,0,0,:,:,:]
        if vis.shape[-1] == 3:
            wgt = vis[:,:,:,2]
            vis = vis[:,:,:,:2]
        else:
            wgt = numpy.ones((vis.shape[0], vis.shape[1], vis.shape[2]), dtype=numpy.float32)
        ## Back to complex
        vis = vis[:,:,:,0] + 1j*vis[:,:,:,1]
        ## Setup a dummy mask
        msk = numpy.zeros(nFreq, dtype=numpy.int16)

        # Re-pack into the data dictionary
        for b in xrange(bl.size):
            if bl[b] >= 65536:
                i = self.standMap[int((bl[b] - 65536) / 2048)]
                j = self.standMap[int((bl[b] - 65536) % 2048)]
            else:
                i = self.standMap[int(bl[b] / 256)]
                j = self.standMap[int(bl[b] % 256)]
            if i == j and not includeAuto:
                ## Skip auto-correlations
                continue
            ri = numpy.where(self.stands == i)[0][0]
            rj = numpy.where(self.stands == j)[0][0]
            
            for p,l in enumerate(self.pols):
                name = NumericStokes[l]
                if len(name) == 2:
                    name = name.lower()
                    
                dataDict['bls'][name].append( (ri,rj) )
                dataDict['uvw'][name].append( uvw[:,b,:] )
                dataDict['vis'][name].append( vis[b,:,p] )
                dataDict['wgt'][name].append( wgt[b,:,p] )
                dataDict['msk'][name].append( msk )
                dataDict['jd' ][name].append( jd[b] )
                
        # Close
        hdulist.close()
        
        # Sort
        if sort:
            sortDataDict(dataDict)
            
        # Prune
        if uvMin != 0 or uvMax != numpy.inf:
            dataDict = pruneBaselineRange(dataDict, uvMin=uvMin, uvMax=uvMax)
            
        # Return
        return dataDict


try:
    from casacore.tables import table
    
    class CorrelatedDataMS(object):
        """
        Class to make accessing information about a MS easy.  This wraps 
        all of the "messy" machinery needed to extract both the metadata and data 
        from the file and return them as common LSL objects.
        
        This class has three main attributes to interact with:
          * getAntennaArray - Return a :class:`lsl.sim.vim.AntennaArray` instance
                              that represents the array where the data was obtained.  
                              This is useful for simulation proposes and computing 
                              source positions.
          * getObserver - Return a ephem.Observer instance representing the array
          * getDataSet - Return a data dictionary of all baselines for a given set
                         of observations
        
        The class also includes a variety of useful metadata attributes:
          * pols - Numpy array of polarization product codes
          * freq - Numpy array of frequency channels in Hz
          * station - LSL :class:`lsl.common.stations.LWAStation` instance for the
                      array
          * dateObs - Datetime object for the reference date of the FIT IDI file
          * antennas - List of :class:`lsl.common.stations.Antenna` instances
        
        .. note::
            The CorrelatedDataMS.antennas attribute should be used over 
            CorrelatedDataMS.station.getAntennas() since the mapping in the MS
            may not be the same as the digitizer order.
        """
        
        def _createEmptyDataDict(self):
            """
            Create an empty data dictionary that is appropriate for the current file.
            """
            
            dataDict = {}
            dataDict['freq'] = self.freq 
            dataDict['isMasked'] = False
            for key in ('uvw', 'vis', 'wgt', 'msk', 'bls', 'jd'):
                dataDict[key] = {}
                
            for p in self.pols:
                name = NumericStokesMS[p]
                if len(name) == 2:
                    name = name.lower()
                for key in ('uvw', 'vis', 'wgt', 'msk', 'bls', 'jd'):
                    dataDict[key][name] = []
                    
            return dataDict
        
        def __init__(self, filename):
            """
            Initialize a new CorrelatedData instance from a MS and fill 
            in the metadata.
            """
            
            if not os.path.isdir(filename) and tarfile.is_tarfile(filename):
                # LASI generate compressed tarballs that contain the MS.  Deal with 
                # those in a transparent manner by automatically unpacking them
                tempdir = tempfile.mkdtemp(prefix='CorrelatedMS-')
                tf = tarfile.open(filename, mode='r:*')
                tf.extractall(tempdir)
                tf.close()
                
                # Find the first directory that could be a MS
                filename = None
                for path in os.listdir(tempdir):
                    path = os.path.join(tempdir, path)
                    if os.path.isdir(path):
                        filename = path
                        break
                        
                # Clean up the temporary directory when the script exists
                atexit.register(lambda: shutil.rmtree(tempdir))
                
            self.filename = filename
            
            # Open the various tables that we need
            data = table(self.filename, ack=False)
            try:
                ants = table(os.path.join(self.filename, 'ANTENNA'), ack=False)
            except:
                raise RuntimeError("Cannot find table 'ANTENNA' in '%s'" % self.filename)
            try:
                pols = table(os.path.join(self.filename, 'POLARIZATION'), ack=False)
            except:
                raise RuntimeError("Cannot find table 'POLARIZATION' in '%s'" % self.filename)
            try:
                obs = table(os.path.join(self.filename, 'OBSERVATION'), ack=False)
            except:
                raise RuntimeError("Cannot find table 'OBSERVATION' in '%s'" % self.filename)
            try:
                spw = table(os.path.join(self.filename, 'SPECTRAL_WINDOW'), ack=False)
            except:
                raise RuntimeError("Cannot find table 'SPECTRAL_WINDOW' in '%s'" % self.filename)
            try:
                dsc = table(os.path.join(self.filename, 'DATA_DESCRIPTION'), ack=False)
            except:
                raise RuntimeError("Cannot find table 'DATA_DESCRIPTION' in '%s'" % self.filename)
                
            # Station/telescope information
            self.telescope = obs.col('TELESCOPE_NAME')[0]
            
            ## Get latitude and longitude for all antennas
            lat = numpy.array([], dtype=numpy.float64)
            lng = numpy.array([], dtype=numpy.float64)
            elv = numpy.array([], dtype=numpy.float64)
            for row in ants.col('POSITION'):
                la,ln,el = stations.ecef2geo(*row)
                lat = numpy.append(lat, la*180/numpy.pi)
                lng = numpy.append(lng, ln*180/numpy.pi)
                elv = numpy.append(elv, el)
                
            ## Estimate the center of the station
            sLat = robust.mean(lat)
            sLng = robust.mean(lng)
            sElv = robust.mean(elv)
            
            ## Build a preliminayr represenation of the station
            self.station = stations.LWAStation(ants.col('STATION')[0], sLat, sLng, sElv)
            
            ## Fill in the antennas instances
            antennas = []
            for i in xrange(lat.size):
                enz = self.station.getENZOffset((lat[i], lng[i], elv[i]))
                sid = int(ants.col('NAME')[i].translate(None, 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'))
                
                stand = stations.Stand(sid, *enz)
                antennas.append( stations.Antenna(2*(stand.id-1)-1, stand=stand) )
            self.station.antennas = antennas
            
            # Antennas
            self.standMap = {}
            self.stands = []
            for i,noact in enumerate(ants.col('NAME')):
                noact = int(noact[3:])
                self.standMap[i] = noact
                self.stands.append(noact)
                
            self.antennaMap = {}
            self.antennas = []
            for stand in self.stands:
                for ant in self.station.getAntennas():
                    if ant.stand.id == stand and ant.pol == 0:
                        self.antennas.append(ant)
                        self.antennaMap[ant.stand.id] = ant
                        break
                        
            # Polarization and frequency
            self.pols = pols.col('CORR_TYPE')[0]
            self.freq  = numpy.array( spw.col('CHAN_FREQ')[0] )
            
            # Total baseline count
            self.totalBaselineCount = data.nrows()
            
            # Data set times
            self._times = []
            for t in data.col('TIME'):
                if t not in self._times:
                    self._times.append(t)
            jd = self._times[0] / 3600.0 / 24.0 + astro.MJD_OFFSET
            self.dateObs = datetime.utcfromtimestamp(astro.utcjd_to_unix(jd))
            
            # Data set spectral windows
            self._windows = []
            for d in dsc.col('SPECTRAL_WINDOW_ID'):
                self._windows.append( d )
                
            # Integration count
            jd = numpy.array(self._times) / 3600.0 / 24.0 + astro.MJD_OFFSET
            self.integrationCount = len(numpy.unique(jd))
            
            # Close
            data.close()
            ants.close()
            pols.close()
            obs.close()
            spw.close()
            dsc.close()
            
            # Catch multi-spectral window data sets
            if len(self._windows) > 1:
                raise RuntimeError("MeasurementSet with more than one spectral window are not supported")
                
        def getAntennaArray(self):
            """
            Return an AIPY AntennaArray instance for the array that made the 
            observations contained here.
            """
            
            # Get the date of observations
            refJD = astro.unix_to_utcjd(timegm(self.dateObs.timetuple()))
            
            # Return
            return simVis.buildSimArray(self.station, self.antennas, self.freq/1e9, jd=refJD)
            
        def getObserver(self):
            """
            Return a ephem.Observer instances for the array described in the file.
            """
            
            # Get the date of observations
            refJD = astro.unix_to_utcjd(timegm(self.dateObs.timetuple()))
            
            obs = self.station.getObserver()
            obs.date = refJD - astro.DJD_OFFSET
            return obs
            
        def getDataSet(self, set, includeAuto=False, sort=True, uvMin=0, uvMax=numpy.inf):
            """
            Return a baseline sorted data dictionary for the specified data set.  
            By default this excludes the autocorrelations.  To include 
            autocorrelations set the value of 'includeAuto' to True.  Setting the
            'sort' keyword to False will disable the baseline sorting.  Optionally,
            baselines with lengths between uvMin and uvMax can only be returned.

            .. note::
                uvMin and uvMax should be specified in lambda
                
            .. versionchanged:: 1.1.0
                'set' can now be either an integer or a list to pull back multiple 
                integrations.
            """
            
            # Open the data table
            data = table(self.filename, ack=False)
            
            # We need this a lot...
            nPol = len(self.pols)
            
            # Define the dictionary to return
            dataDict = self._createEmptyDataDict()
            
            # Load in something we can iterate over
            uvw  = data.col('UVW')
            ant1 = data.col('ANTENNA1')
            ant2 = data.col('ANTENNA2')
            vis  = data.col('DATA')
            time = data.col('TIME')
            
            # Set the time to look for
            targetTime = self._times[set-1]
            
            # Loop over data rows
            found = False
            for u,a1,a2,v,t in zip(uvw, ant1, ant2, vis, time):
                if t != targetTime:
                    continue
                found = True

                if a1 == a2 and not includeAuto:
                    ## Skip auto-correlations
                    continue
                r1 = a1
                r2 = a2
                   
                jd = t / 3600.0 / 24.0 + astro.MJD_OFFSET
                u = numpy.array(u)
                v = numpy.array(v)
                w = numpy.ones(v.shape)
                m = numpy.zeros(v.shape, dtype=numpy.int16)
                
                u2 = numpy.zeros((u.size, v.shape[0]), dtype=u.dtype)
                for c in xrange(v.shape[0]):
                    u2[:,c] = u * (self.freq[c] / vLight)
                    
                for c,p in enumerate(self.pols):
                    name = NumericStokesMS[p]
                    if len(name) == 2:
                        name = name.lower()
                        
                    dataDict['uvw'][name].append( u2 ) 
                    dataDict['vis'][name].append( v[:,c] )
                    dataDict['wgt'][name].append( w[:,c] )
                    dataDict['msk'][name].append( m[:,c] )
                    dataDict['bls'][name].append( (r1,r2) )
                    dataDict['jd' ][name].append( jd )
            data.close()
            
            # Make sure we found something
            if not found:
                raise RuntimeError("Cannot find baseline set %i in MS", set)
            
            # Sort
            if sort:
                sortDataDict(dataDict)
                
            # Prune
            if uvMin != 0 or uvMax != numpy.inf:
                dataDict = pruneBaselineRange(dataDict, uvMin=uvMin, uvMax=uvMax)
                
            # Return
            return dataDict
            
except ImportError:
    import warnings
    warnings.warn('Cannot import casacore.tables, MS support disabled', ImportWarning)
    
    class CorrelatedMS(object):
        """
        Class to make accessing information about a MS easy.  This wraps 
        all of the "messy" machinery needed to extract both the metadata and data 
        from the file and return them as common LSL objects.
        """
        
        def __init__(self, filename):
            raise RuntimeError("Cannot import casacore.tables, MS support disabled")


class ImgWPlus(aipy.img.ImgW):
    """
    Sub-class of the aipy.img.ImgW class that adds support for different 
    visibility weighting scheme and uv plane tapering.  This class also
    adds in a couple of additional methods that help determine the size of
    the field of view and the pixels near the phase center.
    """
    
    def __init__(self, size=100, res=1, wres=.5, mf_order=0):
        """size = number of wavelengths which the UV matrix spans (this 
        determines the image resolution).
        res = resolution of the UV matrix (determines image field of view).
        wres: the gridding resolution of sqrt(w) when projecting to w=0."""
        self.res = float(res)
        self.size = float(size)
        ## Small change needed to work with Numpy 1.12+
        dim = numpy.int64(numpy.round(self.size / self.res))
        self.shape = (dim,dim)
        self.uv = numpy.zeros(shape=self.shape, dtype=numpy.complex64)
        self.bm = []
        for i in range(mf_order+1):
            self.bm.append(numpy.zeros(shape=self.shape, dtype=numpy.complex64))
        self.wres = float(wres)
        self.wcache = {}
    
    def put(self, uvw, data, wgts=None, invker2=None):
        """Same as Img.put, only now the w component is projected to the w=0
        plane before applying the data to the UV matrix."""
        u, v, w = uvw
        if len(u) == 0: return
        if wgts is None:
            wgts = []
            for i in range(len(self.bm)):
                if i == 0:
                    wgts.append(numpy.ones_like(data))
                else:
                    wgts.append(numpy.zeros_like(data))
        if len(self.bm) == 1 and len(wgts) != 1:
            wgts = [wgts]
        assert(len(wgts) == len(self.bm))
        # Sort uvw in order of w
        order = numpy.argsort(w)
        u = u.take(order)
        v = v.take(order)
        w = w.take(order)
        data = data.take(order)
        wgts = [wgt.take(order) for wgt in wgts]
        sqrt_w = numpy.sqrt(numpy.abs(w)) * numpy.sign(w)
        i = 0
        while True:
            # Grab a chunk of uvw's that grid w to same point.
            j = sqrt_w.searchsorted(sqrt_w[i]+self.wres)
            print('%d/%d datums' % (j, len(w)))
            avg_w = numpy.average(w[i:j])
            # Put all uv's down on plane for this gridded w point
            wgtsij = [wgt[i:j] for wgt in wgts]
            uv,bm = aipy.img.Img.put(self, (u[i:j],v[i:j],w[i:j]),
                data[i:j], wgtsij, apply=False)
            # Convolve with the W projection kernel
            invker = numpy.fromfunction(lambda u,v: self.conv_invker(u,v,avg_w),
                uv.shape)
            if not invker2 is None:
                invker *= invker2
            self.uv += ifft2Function(fft2Function(uv) * invker)
            #self.uv += uv
            for b in range(len(self.bm)):
                self.bm[b] += ifft2Function(fft2Function(bm[b]) * invker)
                #self.bm[b] += numpy.array(bm)[0,:,:]
            if j >= len(w):
                break
            i = j
            
    def getFieldOfView(self):
        """
        Return the approximate size of the field of view in radians.  The 
        field of view calculate is based off the maximum and minimum values
        of L found for the inverted uv matrix.
        """
        
        # Get the L and M coordinates
        l,m = self.get_LM()
        
        # Find the maximum and minimum values of L
        lMax = numpy.where( l == l.max() )
        lMin = numpy.where( l == l.min() )
        #print lMax, lMin, l[lMax], l[lMin], m[lMax], m[lMin]
        
        # Convert these locations into topocentric
        xMax, xMin = l.data[lMax], l.data[lMin]
        yMax, yMin = m.data[lMax], m.data[lMin]
        zMax, zMin = numpy.sqrt(1 - xMax**2 - yMax**2), numpy.sqrt(1 - xMin**2 - yMin**2)
        azAltMax = aipy.coord.top2azalt((xMax,yMax,zMax))
        azAltMin = aipy.coord.top2azalt((xMin,yMin,zMin))
        
        # Get the separation between the two
        d = 2*numpy.arcsin( numpy.sqrt( numpy.sin((azAltMax[1]-azAltMin[1])/2)**2+numpy.cos(azAltMax[1])*numpy.cos(azAltMin[1])*numpy.sin((azAltMax[0]-azAltMin[0])/2)**2 ) )
        
        return d.max()
        
    def getPixelSize(self):
        """
        Return the approximate size of pixels at the phase center in radians.
        The pixel size is averaged over the four pixels that neighboor the 
        phase center.
        """
        
        # Get the L and M coordinates
        l,m = self.get_LM()
        
        sizes = []
        x0, y0 = l[0,0], m[0,0]
        z0 = numpy.sqrt(1 - x0**2 - y0**2)
        for offX,offY in ((0,1), (1,0), (0,-1), (-1,0)):
            x1, y1 = l[offX,offY], m[offX,offY]
            z1 = numpy.sqrt(1 - x1**2 - y1**2)
            
            # Convert these locations into topocentric
            azAlt0 = aipy.coord.top2azalt((x0,y0,z0))
            azAlt1 = aipy.coord.top2azalt((x1,y1,z1))
            
            # Get the separation between the two
            d = 2*numpy.arcsin( numpy.sqrt( numpy.sin((azAlt1[1]-azAlt0[1])/2)**2+numpy.cos(azAlt0[1])*numpy.cos(azAlt1[1])*numpy.sin((azAlt1[0]-azAlt0[0])/2)**2 ) )
            
            # Save
            sizes.append(d)
        sizes = numpy.array(sizes)
        
        return sizes.mean()
        
    def _gen_img(self, data, center=(0,0), weighting='natural', localFraction=0.5, robust=0.0, taper=(0.0, 0.0)):
        """
        Return the inverse FFT of the provided data, with the 0,0 point 
        moved to 'center'.  In the images return north is up and east is 
        to the left.
        
        There are a few keywords that control how the image is formed.  
        There are:
          * weighting - The weighting scheme ('natural', 'uniform', or 
                       'briggs') used on the data;
          * localFraction - The fraction of the uv grid that is consider 
                            "local" for the 'uniform' and 'briggs' methods;
          * robust - The value for the weighting robustness under the 
                     'briggs' method; and
          * taper - The size of u and v Gaussian tapers at the 30% level.
        """
        
        # Make sure that we have a valid weighting scheme to use
        if weighting not in ('natural', 'uniform', 'briggs'):
            raise ValueError("Unknown weighting scheme '%s'" % weighting)
            
        # Make sure that we have a valid localFraction value
        if localFraction <= 0 or localFraction > 1:
            raise ValueError("Invalid localFraction value")
            
        # Apply the weighting
        if weighting == 'natural':
            ## Natural weighting - we already have it
            pass
        
        elif weighting == 'uniform':
            ## Uniform weighting - we need to calculate it
            dens = numpy.abs(self.bm[0])
            size = dens.shape[0]
            
            from scipy.ndimage import uniform_filter
            dens = uniform_filter(dens, size=size*localFraction)
            dens /= dens.max()
            dens[numpy.where( dens < 1e-8 )] = 0
            
            data = data/dens
            data[numpy.where(dens == 0)] = 0.0
            
        elif weighting == 'briggs':
            ## Robust weighting - we need to calculate it
            dens = numpy.abs(self.bm[0])
            size = dens.shape[0]
            
            from scipy.ndimage import uniform_filter
            dens = uniform_filter(dens, size=size*localFraction)
            dens /= dens.max()
            dens[numpy.where( dens < 1e-8 )] = 0
            
            f2 = (5*10**-robust)**2 / (dens**2).mean()
            dens = 1.0 / (1.0 + f2/dens)
            data = data/dens*dens.max()
            data[numpy.where(dens == 0)] = 0.0
            
        # Make sure that we have the right type to taper with
        try:
            taper1 = taper[0]
            taper2 = taper[1]
            taper = (taper1, taper2)
        except TypeError:
            taper = (taper, taper)
            
        # Apply the taper
        if taper[0] > 0.0 or taper[1] > 0.0:
            u,v = self.get_uv()
            
            taper1 = 1.0
            if taper[0] > 0.0:
                cu = numpy.log(0.3) / taper[0]**2
                taper1 = numpy.exp(cu*u**2)
                
            taper2 = 1.0
            if taper[1] > 0.0:
                cv = numpy.log(0.3) / taper[1]**2
                taper2 = numpy.exp(cv*v**2)
                
            data = data*taper1*taper2
            
        return aipy.img.recenter(ifft2Function(data).real.astype(numpy.float32), center)
        
    def image(self, center=(0,0), weighting='natural', localFraction=0.5, robust=0.0, taper=(0.0, 0.0)):
        """Return the inverse FFT of the UV matrix, with the 0,0 point moved
        to 'center'.  In the images return north is up and east is 
        to the left.
        
        There are a few keywords that control how the image is formed.  
        There are:
          * weighting - The weighting scheme ('natural', 'uniform', or 
                       'briggs') used on the data;
          * localFraction - The fraction of the uv grid that is consider 
                            "local" for the 'uniform' and 'briggs' methods;
          * robust - The value for the weighting robustness under the 
                     'briggs' method; and
          * taper - The size of u and v Gaussian tapers at the 30% level.
        """
        
        return self._gen_img(self.uv, center=center, weighting=weighting, localFraction=localFraction, robust=robust, taper=taper)
        
    def bm_image(self, center=(0,0), term=None, weighting='natural', localFraction=0.5, robust=0.0, taper=(0.0, 0.0)):
        """Return the inverse FFT of the sample weightings (for all mf_order
        terms, or the specified term if supplied), with the 0,0 point
        moved to 'center'.  In the images return north is up and east is 
        to the left.
        
        There are a few keywords that control how the image is formed.  
        There are:
          * weighting - The weighting scheme ('natural', 'uniform', or 
                       'briggs') used on the data;
          * localFraction - The fraction of the uv grid that is consider 
                            "local" for the 'uniform' and 'briggs' methods;
          * robust - The value for the weighting robustness under the 
                     'briggs' method; and
          * taper - The size of u and v Gaussian tapers at the 30% level.
        """
        
        if not term is None:
            return self._gen_img(self.bm[term], center=center, weighting=weighting, localFraction=localFraction, robust=robust, taper=taper)
        else:
            return [self._gen_img(b, center=center, weighting=weighting, localFraction=localFraction, robust=robust, taper=taper) for b in self.bm]


def buildGriddedImage(dataDict, MapSize=80, MapRes=0.50, MapWRes=0.10, pol='xx', chan=None, verbose=True):
    """
    Given a data dictionary, build an aipy.img.ImgW object of gridded uv data 
    which can be used for imaging.  The ImgW object itself is returned by this 
    function to make it more versatile.
    """
    
    im = ImgWPlus(size=MapSize, res=MapRes, wres=MapWRes)
    
    # Make sure we have the right polarization
    if pol not in dataDict['bls'].keys() and pol.lower() not in dataDict['bls'].keys():
        raise RuntimeError("Data dictionary does not have data for polarization '%s'" % pol)
        
    if chan is not None:
        # Make sure that `chan' is an array by trying to find its length
        try:
            junk = len(chan)
        except TypeError:
            chan = [chan]
            
        # Build up the data using only the specified channels
        uvw = []
        vis = []
        wgt = []
        for (i,j),d,u,w,m in zip(dataDict['bls'][pol], dataDict['vis'][pol], dataDict['uvw'][pol], dataDict['wgt'][pol], dataDict['msk'][pol]):
            u = u[:,chan]
            u.shape = (3, len(chan))
            
            uvw.append(u)
            vis.append(numpy.array([d[chan]]))
            wgt.append(numpy.array([w[chan]]))
    else:
        uvw = dataDict['uvw'][pol]
        vis = dataDict['vis'][pol]
        wgt = dataDict['wgt'][pol]
        
    uvw = numpy.concatenate(uvw, axis=1)
    vis = numpy.concatenate(vis)
    wgt = numpy.concatenate(wgt)
    
    if not verbose:
        sys.stdout = StringIO.StringIO()
        
    uvw, vis, wgt = im.append_hermitian(uvw, vis, wgts=wgt)
    im.put(uvw, vis, wgts=wgt)
    
    if not verbose:
        sys.stdout.close()
        sys.stdout = sys.__stdout__
        
    return im


def plotGriddedImage(ax, gimg, shifted=True, origin='lower', interpolation='nearest', **kwargs):
    """
    Given a blank matplotlib axes instance and a gridded image generated by 
    the buildGriddedImage() function, plot the image on the axes and setup
    the basic coordinate system.  This function returns the matplotlib object
    added to the plot
    
    .. versionchanged:: 1.2.1
        Changed the function to return the matplotlib object plotted so
        that colorbars can be added
        
    .. versionchanged:: 1.1.0
        Added a 'shifted' keyword to control whether or not the image
        is centered or not.
        
    .. versionadded:: 1.0.1
    """
    
    # Build the unshifted image
    img = gimg.image()
    
    # Shift the image so that it is centered in the frame
    if shifted:
        imgSize = img.shape[0]	# should be square
        img = numpy.roll(img, imgSize/2, axis=0)
        img = numpy.roll(img, imgSize/2, axis=1)
        
    # Get the extent in L and M
    l, m = gimg.get_LM()
    lmin, lmax = l.min(), l.max()
    mmin, mmax = m.min(), m.max()
    extent = (mmax,mmin, lmin,lmax)
    if origin != 'lower':
        extent = (mmin,mmax, lmin,lmax)
        
    # Plot
    return ax.imshow(img, extent=extent, origin=origin, interpolation=interpolation, **kwargs)


def getImageRADec(gimg, aa, phaseCenter='z', shifted=True):
    """
    Given a gridded image generated by the buildGriddedImage() function
    and an AntennaArray instance, return a two-element tuple containing
    the RA and dec. values (in radians) for each pixel in the image.  
    
    The 'phaseCenter' keyword controls what the phase center of the image 
    is and defaults to zenith.
    
    .. versionadded: 1.1.0
    """
    
    # Get the phase center
    if phaseCenter is not 'z':
        phaseCenter.compute(aa)
        pcRA, pcDec = phaseCenter.ra, phaseCenter.dec
    else:
        pcRA, pcDec = aa.sidereal_time(), aa.lat
    rotInv = aipy.coord.top2eq_m(0, pcDec)
    
    # Extract the raw topocentric coordinates and convert to equatorial
    top = gimg.get_top()
    oldShape = top[0].shape
    top = (top[0].ravel(), top[1].ravel(), top[2].ravel())
    eq = numpy.dot(rotInv, top)
    eq = (eq[0].reshape(oldShape), eq[1].reshape(oldShape), eq[2].reshape(oldShape))
    
    # Over to RA/Dec
    ra, dec = aipy.coord.eq2radec(eq)
    
    # Correct for the phaseCenter
    ra += pcRA
    ra %= 2*numpy.pi
    
    # Shift, if needed
    if shifted:
        raSize = ra.shape[0]	# should be square
        ra = numpy.roll(ra, raSize/2, axis=0)
        ra = numpy.roll(ra, raSize/2, axis=1)
        
        decSize = dec.shape[0]	# should be square
        dec = numpy.roll(dec, decSize/2, axis=0)
        dec = numpy.roll(dec, decSize/2, axis=1)
        
    # Done
    return ra, dec


def getImageAzEl(gimg, aa, phaseCenter='z', shifted=True):
    """
    Given a gridded image generated by the buildGriddedImage() function
    and an AntennaArray instance, return a two-element tuple containing
    the azimuth and elevation (altitude), both in radians, for each pixel
    in the image.
    
    The 'phaseCenter' keyword controls what the phase center of the image 
    is and defaults to zenith.
    
    .. versionadded: 1.1.0
    """
    
    # Get the phase center
    if phaseCenter is not 'z':
        phaseCenter.compute(aa)
        pcRA, pcDec = phaseCenter.ra, phaseCenter.dec
    else:
        pcRA, pcDec = aa.sidereal_time(), aa.lat
    rot = aipy.coord.eq2top_m(0, pcDec)
    
    # Get the RA and dec. coordinates for each pixel
    ra, dec = getImageRADec(gimg, aa, phaseCenter=phaseCenter, shifted=shifted)
    
    # Convert to azimuth and elevation using PyEphem
    bdy = aipy.amp.RadioFixedBody(0, 0)
    
    az, el = ra*0.0, dec*0.0
    for i in xrange(az.shape[0]):
        for j in xrange(az.shape[1]):
            bdy._ra = ra[i,j]
            bdy._dec = dec[i,j]
            bdy.compute(aa)
            az[i,j], el[i,j] = bdy.az, bdy.alt
    
    # Done
    return az, el



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function
import sys
if sys.version_info > (3,):
    xrange = range
    long = int

"""
Python module for creating creating, validating, and writing simulated 
TBW frames to a file.
"""

import numpy

from lsl.common.dp import fS
from lsl.reader import tbw
from lsl.sim.errors import invalidStand, invalidDataSize, invalidDataType

__version__ = '0.2'
__revision__ = '$Rev: 2455 $'
__all__ = ['SimFrame', 'frame2frame', '__version__', '__revision__', '__all__']


import warnings
warnings.warn("lsl.sim.tbw will be removed in LSL version 1.3.0", DeprecationWarning)


def frame2frame(tbwFrame):
    """
    Convert a :class:`lsl.reader.tbw.Frame` object to a raw DP TBW frame.
    """
    
    # The raw frame
    rawFrame = numpy.zeros(tbw.FrameSize, dtype=numpy.uint8)

    # Part 1: The header
    ## Sync. words (0xDEC0DE5C)
    rawFrame[0] = 0xDE  # 222
    rawFrame[1] = 0xC0  # 192
    rawFrame[2] = 0xDE  # 222
    rawFrame[3] = 0x5C  #  92
    ## Frame count
    rawFrame[5] = (tbwFrame.header.frameCount>>16) & 255
    rawFrame[6] = (tbwFrame.header.frameCount>>8) & 255
    rawFrame[7] = tbwFrame.header.frameCount & 255
    ## Seconds count
    rawFrame[8] = (tbwFrame.header.secondsCount>>24) & 255
    rawFrame[9] = (tbwFrame.header.secondsCount>>16) & 255
    rawFrame[10] = (tbwFrame.header.secondsCount>>8) & 255
    rawFrame[11] = tbwFrame.header.secondsCount & 255
    ## TBW ID
    rawFrame[12] = (tbwFrame.header.tbwID>>8) & 255
    rawFrame[13] = tbwFrame.header.tbwID & 255
    ## NB: Next two bytes are unsigned
    
    # Part 2: The data
    ## Time tag
    rawFrame[16] = (tbwFrame.data.timeTag>>56) & 255
    rawFrame[17] = (tbwFrame.data.timeTag>>48) & 255
    rawFrame[18] = (tbwFrame.data.timeTag>>40) & 255
    rawFrame[19] = (tbwFrame.data.timeTag>>32) & 255
    rawFrame[20] = (tbwFrame.data.timeTag>>24) & 255
    rawFrame[21] = (tbwFrame.data.timeTag>>16) & 255
    rawFrame[22] = (tbwFrame.data.timeTag>>8) & 255
    rawFrame[23] = tbwFrame.data.timeTag & 255
    ## Data - common
    x = numpy.squeeze(tbwFrame.data.xy[0,:])
    y = numpy.squeeze(tbwFrame.data.xy[1,:])
    ## Data - 12 bit
    if tbwFrame.getDataBits() == 12:
        ### Round, clip, and convert to unsigned integers
        x = x.round()
        x = x.clip(-2048, 2047)
        x = x.astype(numpy.uint16)
        y = y.round()
        y = y.clip(-2048, 2047)
        y = y.astype(numpy.uint16)
        
        rawFrame[24::3] = (x >> 4) & 255
        rawFrame[25::3] = ((x & 15) << 4) | ((y >> 8) & 15)
        rawFrame[26::3] = y & 255
        
    ## Data - 4 bit
    if tbwFrame.getDataBits() == 4:
        ### Round, clip, and convert to unsigned integers
        x = x.round()
        x = x.clip(-8, 7)
        x = x.astype(numpy.uint8)
        y = y.round()
        y = y.clip(-8, 7)
        y = y.astype(numpy.uint8)

        rawFrame[24:] = (x << 4) | y
    
    return rawFrame


class SimFrame(tbw.Frame):
    """
    tbw.SimFrame extends the :class:`lsl.reader.tbw.Frame` object to yield a method 
    for easily creating DP ICD-compliant raw TBW frames.  Frames created with
    this method can be written to a file via the methods writeRawFrame() function.
    """

    def __init__(self, stand=None, frameCount=None, dataBits=12, obsTime=None, xy=None):
        """
        Given a list of parameters, build a tbw.SimFrame object.  The parameters
        needed are:
        * stand id (>0 & <259)
        * which frame number to create
        * dataBits (12 or 4)
        * observation time in samples at fS since the epoch
        * 2-D numpy array representing the frame data for both polarizations.
        
        Not all of these parameters are needed at initialization of the object and
        the values can be added later.

        .. versionchanged:: 0.3.4
            obsTime now in samples at fS, not seconds
        """
        
        self.stand = stand
        self.frameCount = frameCount
        self.dataBits = dataBits
        self.obsTime = obsTime
        self.xy = xy
        super(SimFrame, self).__init__()
        
    def __update(self):
        """
        Private function to use the object's parameter values to build up 
        a tbw.Frame-like object.
        """
        
        self.header.frameCount = self.frameCount
        self.header.secondsCount = long(self.obsTime / fS)
        if self.dataBits == 12:
            self.header.tbwID = 32768 | self.stand
        else:
            self.header.tbwID = 32768 | 16384 | self.stand
        
        self.data.timeTag = self.obsTime
        self.data.xy = self.xy
        
    def loadFrame(self, tbwFrame):
        """
        Populate the a tbw.SimFrame object with a pre-made frame.
        """
        
        self.header = tbwFrame.header
        self.data = tbwFrame.data
        
        # Back-fill the class' fields to make sure the object is consistent
        ## Header
        self.stand = self.header.parseID()
        self.frameCount = self.header.frameCount
        self.dataBits = self.header.getDataBits()
        ## Data
        self.obsTime = self.data.timeTag
        self.xy = self.data.xy
    
    def isValid(self, raiseErrors=False):
        """
        Check if simulated TBW frame is valid or not.  Valid frames return 
        True and invalid frames False.  If the 'raiseErrors' keyword is set, 
        isValid() raises an error when a problem with the frame structure is 
        encountered.
        """

        # Make sure we have the latest values
        self.__update()

        stand = self.parseID()
        # Is the stand number reasonable?
        if stand == 0 or stand > 258:
            if raiseErrors:
                raise invalidStand()
            return False

        # Is there data loaded into frame.data.xy?
        if self.data.xy is None:
            if raiseErrors:
                raise invalidDataSize()
            return False

        # Does the data shape make sense?
        if self.data.xy.shape[0] != 2:
            if raiseErrors:
                raise invalidDataSize()
            return False

        # Does the data length match the data bits in the header?
        if self.getDataBits() == 12 and self.data.xy.shape[1] != 400:
            if raiseErrors:
                raise invalidDataSize()
            return False
        if self.getDataBits() == 4 and self.data.xy.shape[1] != 1200:
            if raiseErrors:
                raise invalidDataSize()
            return False

        # Does the data type make sense?
        if self.data.xy.dtype.kind == 'c':
            if raiseErrors:
                raise invalidDataType()
            return False

        # If we made it this far, it's valid
        return True

    def createRawFrame(self):
        """
        Re-express a simulated TBW frame as a numpy array of unsigned 8-bit 
        integers.  Returns a numpy array if the frame  is valid.  If the frame 
        is not ICD-compliant, a errors.baseSimError-type error is raised.
        """

        # Make sure we have the latest values
        self.__update()

        self.isValid(raiseErrors=True)
        return frame2frame(self)

    def writeRawFrame(self, fh):
        """
        Write a simulated TBW frame to a filehandle if the frame is valid.
        If the frame is not ICD-compliant, a errors.baseSimError-type error is 
        raised.
        """

        rawFrame = self.createRawFrame()
        rawFrame.tofile(fh)

    def __str__(self):
        if self.stand is None:
            return "Empty TBW SimFrame object"
        else:
            return "%i-bit TBW SimFrame for stand %i @ time %i" % (self.dataBits, self.stand, self.obsTime)



# -*- coding: utf-8 -*-

"""
Unit test for the lsl.transform module.
"""

import os
import math
import ephem
import unittest

from lsl import transform
from lsl.astro import DJD_OFFSET
from lsl.common.stations import lwa1


__revision__ = "$Rev: 2394 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


class transform_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.transform
    module."""
    
    def test_time_init(self):
        """Test the transform.Time constructor."""
        
        t0 = transform.Time('2013-01-08 01:23:45.000', format='STR')
        self.assertEqual(t0.utc_str, '2013-01-08 01:23:45.000')
        
    def test_planetaryposition_init(self):
        """Test the transform.PlanetaryPosition constructor."""
        
        p0 = transform.PlanetaryPosition('Jupiter')
        p1 = transform.PlanetaryPosition('Sun')
        
    def test_planetaryposition_saturn(self):
        """Test the location of Saturn."""
        
        t0 = transform.Time('2013-01-08 01:23:45.000', format='STR')
        
        obs = lwa1.getObserver()
        obs.date = t0.utc_str
        sat = ephem.Saturn()
        sat.compute(obs)
        
        p0 = transform.PlanetaryPosition('Saturn')
        
        self.assertAlmostEqual(p0.apparent_equ(t0)[0], sat.g_ra *180.0/math.pi, 4)
        self.assertAlmostEqual(p0.apparent_equ(t0)[1], sat.g_dec*180.0/math.pi, 4)
        
    def test_planetaryposition_jupiter(self):
        """Test the location of Jupiter."""
        
        t0 = transform.Time('2013-01-08 01:23:45.000', format='STR')
        
        obs = lwa1.getObserver()
        obs.date = t0.utc_str
        jove = ephem.Jupiter()
        jove.compute(obs)
        
        p0 = transform.PlanetaryPosition('Jupiter')
        
        self.assertAlmostEqual(p0.apparent_equ(t0)[0], jove.g_ra *180.0/math.pi, 4)
        self.assertAlmostEqual(p0.apparent_equ(t0)[1], jove.g_dec*180.0/math.pi, 4)
        
    def test_planetaryposition_mars(self):
        """Test the location of Mars."""
        
        t0 = transform.Time('2013-01-08 01:23:45.000', format='STR')
        
        obs = lwa1.getObserver()
        obs.date = t0.utc_str
        mars = ephem.Mars()
        mars.compute(obs)
        
        p0 = transform.PlanetaryPosition('Mars')
        
        self.assertAlmostEqual(p0.apparent_equ(t0)[0], mars.g_ra *180.0/math.pi, 4)
        self.assertAlmostEqual(p0.apparent_equ(t0)[1], mars.g_dec*180.0/math.pi, 4)
        
    def test_planetaryposition_venus(self):
        """Test the location of Venus."""
        
        t0 = transform.Time('2013-01-08 01:23:45.000', format='STR')
        
        obs = lwa1.getObserver()
        obs.date = t0.utc_str
        venu = ephem.Venus()
        venu.compute(obs)
        
        p0 = transform.PlanetaryPosition('Venus')
        
        self.assertAlmostEqual(p0.apparent_equ(t0)[0], venu.g_ra *180.0/math.pi, 4)
        self.assertAlmostEqual(p0.apparent_equ(t0)[1], venu.g_dec*180.0/math.pi, 4)
        
    def test_planetaryposition_sun(self):
        """Test the location of the Sun."""
        
        t0 = transform.Time('2013-01-08 01:23:45.000', format='STR')
        
        obs = lwa1.getObserver()
        obs.date = t0.utc_str
        sol = ephem.Sun()
        sol.compute(obs)
        
        p0 = transform.PlanetaryPosition('Sun')
        
        self.assertAlmostEqual(p0.apparent_equ(t0)[0], sol.g_ra *180.0/math.pi, 4)
        self.assertAlmostEqual(p0.apparent_equ(t0)[1], sol.g_dec*180.0/math.pi, 4)
        
    def test_geographicalposition_init(self):
        """Test the transform.GeographicalPosition constructor."""
        
        lon = lwa1.long * 180.0/math.pi
        lat = lwa1.lat  * 180.0/math.pi
        elv = lwa1.elev
        
        g0 = transform.GeographicalPosition([lon,lat,elv])
        
    def test_geographicalposition_ecef(self):
        """Test the tranform.GeographicalPosition EC-EF transform."""
        
        lon = lwa1.long * 180.0/math.pi
        lat = lwa1.lat  * 180.0/math.pi
        elv = lwa1.elev
        
        g0 = transform.GeographicalPosition([lon,lat,elv])
        self.assertAlmostEqual(g0.ecef[0], lwa1.getGeocentricLocation()[0], 6)
        self.assertAlmostEqual(g0.ecef[1], lwa1.getGeocentricLocation()[1], 6)
        self.assertAlmostEqual(g0.ecef[2], lwa1.getGeocentricLocation()[2], 6)
        
    def test_geographicalposition_lst(self):
        """Test the tranform.GeographicalPosition sidereal time."""
        
        t0 = transform.Time('2013-01-08 01:23:45.000', format='STR')
        
        lon = lwa1.long * 180.0/math.pi
        lat = lwa1.lat  * 180.0/math.pi
        elv = lwa1.elev
        obs = lwa1.getObserver()
        
        g0 = transform.GeographicalPosition([lon,lat,elv])
        
        # The astro.get_apparent_sidereal_time() function doesn't care about
        # elevation
        obs.elev = 0.0
        
        obs.date = t0.utc_str
        self.assertAlmostEqual(g0.sidereal(t0), obs.sidereal_time()*12.0/math.pi, 4)
        
    def test_pointingdirection_init(self):
        """Test the transform.PointingDirection constructor."""
        
        lon = lwa1.long * 180.0/math.pi
        lat = lwa1.lat  * 180.0/math.pi
        elv = lwa1.elev
        
        g0 = transform.GeographicalPosition([lon,lat,elv])
        p0 = transform.PlanetaryPosition('Jupiter')
        
        d0 = transform.PointingDirection(p0, g0)
        
    def test_pointingdirection_azalt(self):
        """Test the transform.PointingDirection az/alt transform."""
        
        t0 = transform.Time('2013-01-08 01:23:45.000', format='STR')
        
        lon = lwa1.long * 180.0/math.pi
        lat = lwa1.lat  * 180.0/math.pi
        elv = lwa1.elev
        obs = lwa1.getObserver()
        obs.date = t0.utc_str
        jove = ephem.Jupiter()
        jove.compute(obs)
        
        g0 = transform.GeographicalPosition([lon,lat,elv])
        p0 = transform.PlanetaryPosition('Jupiter')
        d0 = transform.PointingDirection(p0, g0)
        
        self.assertAlmostEqual(d0.hrz(t0)[0], jove.az *180.0/math.pi, 1)
        self.assertAlmostEqual(d0.hrz(t0)[1], jove.alt*180.0/math.pi, 1)
        
    def test_pointingdirection_azalt(self):
        """Test the transform.PointingDirection az/alt transform."""
        
        t0 = transform.Time('2013-01-08 01:23:45.000', format='STR')
        
        lon = lwa1.long * 180.0/math.pi
        lat = lwa1.lat  * 180.0/math.pi
        elv = lwa1.elev
        obs = lwa1.getObserver()
        obs.date = t0.utc_str
        jove = ephem.Jupiter()
        jove.compute(obs)
        
        g0 = transform.GeographicalPosition([lon,lat,elv])
        p0 = transform.PlanetaryPosition('Jupiter')
        d0 = transform.PointingDirection(p0, g0)
        
        rst = d0.rst(t0)
        self.assertAlmostEqual(rst.rise, obs.next_rising(jove)*1.0+DJD_OFFSET, 2)
        self.assertAlmostEqual(rst.transit, obs.next_transit(jove)*1.0+DJD_OFFSET, 2)
        self.assertAlmostEqual(rst.set, obs.next_setting(jove)*1.0+DJD_OFFSET, 2)


class transform_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.transform units 
    tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(transform_tests)) 


if __name__ == '__main__':
    unittest.main()
